""" Implements various methods for non-exhaustive template matching
    based on numerical optimization.

    Copyright (c) 2023 European Molecular Biology Laboratory

    Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

from typing import Tuple, Dict
from abc import ABC, abstractmethod

import numpy as np
from numpy.typing import NDArray
from scipy.optimize import (
    differential_evolution,
    LinearConstraint,
    basinhopping,
)
from scipy.ndimage import laplace
from scipy.spatial import KDTree

from .matching_utils import rigid_transform, euler_to_rotationmatrix


class MatchCoordinatesToDensity(ABC):
    """
    A class to template match coordinate sets.

    Parameters
    ----------
    target_coordinates : NDArray
        The coordinates of the target.
    template_coordinates : NDArray
        The coordinates of the template.
    target_weights : NDArray
        The weights of the target.
    template_weights : NDArray
        The weights of the template.
    sampling_rate : NDArray
        The size of the voxel.
    template_mask_coordinates : NDArray, optional
        The coordinates of the template mask. Default is None.
    target_mask_coordinates : NDArray, optional
        The coordinates of the target mask. Default is None.
    **kwargs : dict, optional
        Other keyword arguments.
    """

    def __init__(
        self,
        target_coordinates: NDArray,
        template_coordinates: NDArray,
        target_weights: NDArray,
        template_weights: NDArray,
        sampling_rate: NDArray,
        template_mask_coordinates: NDArray = None,
        target_mask_coordinates: NDArray = None,
        **kwargs,
    ):
        target, _, origin = FitRefinement.array_from_coordinates(
            target_coordinates, target_weights, sampling_rate
        )
        self.target_density = target
        self.target_origin = origin
        self.sampling_rate = sampling_rate

        self.template_weights = template_weights
        self.template_coordinates = template_coordinates
        self.template_coordinates_rotated = np.empty(
            self.template_coordinates.shape, dtype=np.float32
        )

        self.target_mask_density = None
        if target_mask_coordinates is not None:
            target_mask, *_ = FitRefinement.array_from_coordinates(
                coordinates=target_mask_coordinates.astype(np.float32),
                weights=np.ones(target_mask_coordinates.shape[1]),
                shape=self.target_density.shape,
                origin=self.target_origin,
                sampling_rate=self.sampling_rate,
            )
            self.target_mask_density = target_mask

        self.template_mask_coordinates = None
        self.template_mask_coordinates_rotated = None
        if template_mask_coordinates is not None:
            self.template_mask_coordinates = template_mask_coordinates
            self.template_mask_coordinates_rotated = np.empty(
                self.template_mask_coordinates.shape, dtype=np.float32
            )

    def __call__(self, x: NDArray):
        """
        Return the score for a given transformation.

        Parameters
        ----------
        x : NDArray
            The input transformation parameters.

        Returns
        -------
        float
            The negative score from the scoring function.
        """
        translation, rotation = x[:3], x[3:]
        rotation_matrix = euler_to_rotationmatrix(rotation)

        rigid_transform(
            coordinates=self.template_coordinates,
            coordinates_mask=self.template_mask_coordinates,
            rotation_matrix=rotation_matrix,
            translation=translation,
            out=self.template_coordinates_rotated,
            out_mask=self.template_mask_coordinates_rotated,
            use_geometric_center=False,
        )

        mapping = FitRefinement.map_coordinates_to_array(
            coordinates=self.template_coordinates_rotated,
            coordinates_mask=self.template_mask_coordinates_rotated,
            array_origin=self.target_origin,
            array_shape=self.target_density.shape,
            sampling_rate=self.sampling_rate,
        )

        return -self.scoring_function(
            transformed_coordinates=mapping[0],
            transformed_coordinates_mask=mapping[1],
            in_volume=mapping[2],
            in_volume_mask=mapping[3],
        )

    @abstractmethod
    def scoring_function(*args, **kwargs):
        """
        Computes a scoring metric for a given set of coordinates.

        This function is not intended to be called directly, but should rather be
        defined by classes inheriting from :py:class:`MatchCoordinatesToDensity`
        to parse a given file format.
        """


class MatchCoordinatesToCoordinates(ABC):
    """
    A class to template match coordinate sets.

    Parameters
    ----------
    target_coordinates : NDArray
        The coordinates of the target.
    template_coordinates : NDArray
        The coordinates of the template.
    target_weights : NDArray
        The weights of the target.
    template_weights : NDArray
        The weights of the template.
    sampling_rate : NDArray
        The size of the voxel.
    template_mask_coordinates : NDArray, optional
        The coordinates of the template mask. Default is None.
    target_mask_coordinates : NDArray, optional
        The coordinates of the target mask. Default is None.
    **kwargs : dict, optional
        Other keyword arguments.
    """

    def __init__(
        self,
        target_coordinates: NDArray,
        template_coordinates: NDArray,
        target_weights: NDArray,
        template_weights: NDArray,
        template_mask_coordinates: NDArray = None,
        target_mask_coordinates: NDArray = None,
        **kwargs,
    ):
        self.target_weights = target_weights
        self.target_coordinates = target_coordinates

        self.template_weights = template_weights
        self.template_coordinates = template_coordinates
        self.template_coordinates_rotated = np.empty(
            self.template_coordinates.shape, dtype=np.float32
        )
        self.target_mask_coordinates = target_mask_coordinates

        self.template_mask_coordinates = None
        self.template_mask_coordinates_rotated = None
        if template_mask_coordinates is not None:
            self.template_mask_coordinates = template_mask_coordinates
            self.template_mask_coordinates_rotated = np.empty(
                self.template_mask_coordinates.shape, dtype=np.float32
            )

    def __call__(self, x: NDArray):
        """
        Return the score for a given transformation.

        Parameters
        ----------
        x : NDArray
            The input transformation parameters.

        Returns
        -------
        float
            The negative score from the scoring function.
        """
        translation, rotation = x[:3], x[3:]
        rotation_matrix = euler_to_rotationmatrix(rotation)

        rigid_transform(
            coordinates=self.template_coordinates,
            coordinates_mask=self.template_mask_coordinates,
            rotation_matrix=rotation_matrix,
            translation=translation,
            out=self.template_coordinates_rotated,
            out_mask=self.template_mask_coordinates_rotated,
            use_geometric_center=False,
        )

        return -self.scoring_function(
            transformed_coordinates=self.template_coordinates_rotated,
            transformed_coordinates_mask=self.template_mask_coordinates_rotated,
        )

    @abstractmethod
    def scoring_function(*args, **kwargs):
        """
        Computes a scoring metric for a given set of coordinates.

        This function is not intended to be called directly, but should rather be
        defined by classes inheriting from :py:class:`MatchCoordinatesToDensity`
        to parse a given file format.
        """


class CrossCorrelation(MatchCoordinatesToDensity):
    """
    Class representing the Cross-Correlation matching score.

    Cross-Correlation score formula:

    .. math::

        \\text{score} = \\text{target_weights} \\cdot \\text{template_weights}

    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.denominator = 1

    def scoring_function(
        self,
        transformed_coordinates: NDArray,
        transformed_coordinates_mask: NDArray,
        in_volume: NDArray,
        in_volume_mask: NDArray,
    ) -> float:
        """
        Compute the Cross-Correlation score.

        Parameters
        ----------
        transformed_coordinates : NDArray
            Transformed coordinates.
        transformed_coordinates_mask : NDArray
            Mask for the transformed coordinates.
        in_volume : NDArray
            Binary mask indicating which ``transformed_coordinates`` are in the
            target volume.
        in_volume_mask : NDArray
            Binary mask indicating which ``transformed_coordinates`` are in the
            target mask volume.

        Returns
        -------
        float
            The Cross-Correlation score.
        """
        score = np.dot(
            self.target_density[tuple(transformed_coordinates[:, in_volume])],
            self.template_weights[in_volume],
        )
        score /= self.denominator
        return score


class LaplaceCrossCorrelation(CrossCorrelation):
    """
    Class representing the Laplace Cross-Correlation matching score.

    The score is computed like CrossCorrelation, but with Laplace filtered
    weights, indicated by the Laplace operator :math:`\\nabla^{2}`.

    .. math::

        \\text{score} = \\nabla^{2} \\text{target_weights} \\cdot
                        \\nabla^{2} \\text{template_weights}

    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.target_density = laplace(self.target_density)

        arr, positions, _ = FitRefinement.array_from_coordinates(
            self.template_coordinates, self.template_weights, self.sampling_rate
        )
        self.template_weights = laplace(arr)[tuple(positions)]


class NormalizedCrossCorrelation(CrossCorrelation):
    """
    Class representing the Normalized Cross-Correlation matching score.

    The score is computed by normalizing the dot product of `target_weights` and
    `template_weights` with the product of their norms. This normalization ensures
    the score lies between -1 and 1, providing a measure of similarity that's invariant
    to scale.

    .. math::

        \\text{score} = \\frac{\\text{target_weights} \\cdot \\text{template_weights}}
                        {\\text{max(target_norm} \\times \\text{template_norm, eps)}}

    Where:

    .. math::

        \\text{target_norm} = ||\\text{target_weights}||

    .. math::

        \\text{template_norm} = ||\\text{template_weights}||

    Here, :math:`||.||` denotes the L2 (Euclidean) norm.

    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        target_norm = np.linalg.norm(self.target_density[self.target_density != 0])
        template_norm = np.linalg.norm(self.template_weights)
        self.denominator = np.fmax(target_norm * template_norm, np.finfo(float).eps)


class NormalizedCrossCorrelationMean(NormalizedCrossCorrelation):
    """
    Class representing the Mean Normalized Cross-Correlation matching score.

    This class extends the Normalized Cross-Correlation by computing the score
    after subtracting the mean from both `target_weights` and `template_weights`.
    This modification enhances the matching score's sensitivity to patterns
    over flat regions in the data.

    Mathematically, the Mean Normalized Cross-Correlation score is computed as:

    .. math::

        \\text{score} = \\frac{(\\text{target_weights} - \\text{mean(target_weights)})
                        \\cdot (\\text{template_weights} -
                        \\text{mean(template_weights)})}
                        {\\text{max(target_norm} \\times \\text{template_norm, eps)}}

    Where:

    .. math::

        \\text{target_norm} = ||\\text{target_weights} - \\text{mean(target_weights)}||

    .. math::

        \\text{template_norm} = ||\\text{template_weights} -
        \\text{mean(template_weights)}||

    Here, :math:`||.||` denotes the L2 (Euclidean) norm, and :math:`\\text{mean(.)}`
    computes the mean of the respective weights.
    """

    def __init__(self, **kwargs):
        print(kwargs["target_weights"].mean())
        kwargs["target_weights"] -= kwargs["target_weights"].mean()
        kwargs["template_weights"] -= kwargs["template_weights"].mean()
        super().__init__(**kwargs)


class MaskedCrossCorrelation(MatchCoordinatesToDensity):
    """
    Class representing the Masked Cross-Correlation matching score.

    The Masked Cross-Correlation computes the similarity between `target_weights`
    and `template_weights` under respective masks. The score is normalized and lies
    between -1 and 1, providing a measure of similarity even in the presence of
    missing or masked data.

    The formula for the Masked Cross-Correlation is:

    .. math::
        \\text{numerator} = \\text{dot}(\\text{target_weights},
                            \\text{template_weights}) -
                            \\frac{\\text{sum}(\\text{mask_target}) \\times
                            \\text{sum}(\\text{mask_template})}
                                  {\\text{mask_overlap}}

    .. math::
        \\text{denominator1} = \\text{sum}(\\text{mask_target}^2) -
                               \\frac{\\text{sum}(\\text{mask_target})^2}
                                     {\\text{mask_overlap}}

    .. math::
        \\text{denominator2} = \\text{sum}(\\text{mask_template}^2) -
                               \\frac{\\text{sum}(\\text{mask_template})^2}
                                     {\\text{mask_overlap}}

    .. math::
        \\text{denominator} = \\sqrt{\\text{denominator1} \\times \\text{denominator2}}

    .. math::
        \\text{score} = \\frac{\\text{numerator}}{\\text{denominator}}
                        \\text{ if denominator } \\neq 0
                        \\text{ else } 0

    Where:

    -   mask_target and mask_template are binary masks for the target_weights
        and template_weights respectively.

    -   mask_overlap represents the number of overlapping non-zero elements in
        the masks.

    References
    ----------
    .. [1]  Masked FFT registration, Dirk Padfield, CVPR 2010 conference
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def scoring_function(
        self,
        transformed_coordinates: NDArray,
        transformed_coordinates_mask: NDArray,
        in_volume: NDArray,
        in_volume_mask: NDArray,
    ) -> float:
        """
        Compute the Masked Cross-Correlation score.

        Parameters
        ----------
        transformed_coordinates : NDArray
            Transformed coordinates.
        transformed_coordinates_mask : NDArray
            Mask for the transformed coordinates.
        in_volume : NDArray
            Binary mask indicating which ``transformed_coordinates`` are in the
            target volume.
        in_volume_mask : NDArray
            Binary mask indicating which ``transformed_coordinates`` are in the
            target mask volume.

        Returns
        -------
        float
            The Masked Cross-Correlation score.
        """
        mask_overlap = np.sum(
            self.target_mask_density[
                tuple(transformed_coordinates_mask[:, in_volume_mask])
            ],
        )
        mask_overlap = np.fmax(mask_overlap, np.finfo(float).eps)

        mask_target = self.target_density[
            tuple(transformed_coordinates_mask[:, in_volume_mask])
        ]
        denominator1 = np.subtract(
            np.sum(mask_target**2),
            np.divide(np.square(np.sum(mask_target)), mask_overlap),
        )
        mask_template = np.multiply(
            self.template_weights[in_volume],
            self.target_mask_density[tuple(transformed_coordinates[:, in_volume])],
        )
        denominator2 = np.subtract(
            np.sum(mask_template**2),
            np.divide(np.square(np.sum(mask_template)), mask_overlap),
        )

        denominator1 = np.fmax(denominator1, 0.0)
        denominator2 = np.fmax(denominator2, 0.0)
        denominator = np.sqrt(np.multiply(denominator1, denominator2))

        numerator = np.dot(
            self.target_density[tuple(transformed_coordinates[:, in_volume])],
            self.template_weights[in_volume],
        )

        numerator -= np.divide(
            np.multiply(np.sum(mask_target), np.sum(mask_template)), mask_overlap
        )

        if denominator == 0:
            return 0

        score = numerator / denominator
        return score


class PartialLeastSquareDifference(MatchCoordinatesToDensity):
    """
    Class representing the Partial Least Square Difference matching score.

    The Partial Least Square Difference (PLSQ) between the target :math:`f` and the
    template :math:`g` is calculated as:

    .. math::

        \\text{d(f,g)} = \\sum_{i=1}^{n} \\| f(\\mathbf{p}_i) - g(\\mathbf{q}_i) \\|^2

    References
    ----------
    .. [1]  Daven Vasishtan and Maya Topf, "Scoring functions for cryoEM density
            fitting", Journal of Structural Biology, vol. 174, no. 2,
            pp. 333--343, 2011. DOI: https://doi.org/10.1016/j.jsb.2011.01.012
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def scoring_function(
        self,
        transformed_coordinates: NDArray,
        transformed_coordinates_mask: NDArray,
        in_volume: NDArray,
        in_volume_mask: NDArray,
    ) -> float:
        """
        Compute the Partial Least Square Difference score.

        Given the transformed coordinates and their associated mask, this function
        computes the difference between target and template densities.

        Parameters
        ----------
        transformed_coordinates : NDArray
            Transformed coordinates.
        transformed_coordinates_mask : NDArray
            Mask for the transformed coordinates.
        in_volume : NDArray
            Binary mask indicating which ``transformed_coordinates`` are in the
            target volume.
        in_volume_mask : NDArray
            Binary mask indicating which ``transformed_coordinates`` are in the
            target mask volume.

        Returns
        -------
        float
            The negative of the Partial Least Square Difference score.
        """
        score = np.sum(
            np.square(
                np.subtract(
                    self.target_density[tuple(transformed_coordinates[:, in_volume])],
                    self.template_weights[in_volume],
                )
            )
        )
        score += np.sum(np.square(self.template_weights[np.invert(in_volume)]))

        return -score


class Chamfer(MatchCoordinatesToCoordinates):
    """
    Class representing the Chamfer matching score.

    The Chamfer distance is computed as:

    .. math::

        \\text{d(f,g)} = \\frac{1}{|X|} \\sum_{\\mathbf{f}_i \\in X}
        \\inf_{\\mathbf{g} \\in Y} ||\\mathbf{f}_i - \\mathbf{g}||_2

    References
    ----------
    .. [1]  Daven Vasishtan and Maya Topf, "Scoring functions for cryoEM density
            fitting", Journal of Structural Biology, vol. 174, no. 2,
            pp. 333--343, 2011. DOI: https://doi.org/10.1016/j.jsb.2011.01.012
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.target_tree = KDTree(self.target_coordinates.T)

    def scoring_function(
        self,
        transformed_coordinates: NDArray,
        transformed_coordinates_mask: NDArray,
        **kwargs,
    ) -> float:
        """
        Compute the Chamfer distance score.

        Given the transformed coordinates and their associated mask, this function
        calculates the average distance between the rotated template coordinates
        and the nearest target coordinates.

        Parameters
        ----------
        transformed_coordinates : NDArray
            Transformed coordinates.

        Returns
        -------
        float
            The negative of the Chamfer distance score.

        """
        dist, _ = self.target_tree.query(self.template_coordinates_rotated.T)
        score = np.mean(dist)
        return -score


class MutualInformation(MatchCoordinatesToDensity):
    """
    Class representing the Mutual Information matching score.

    The Mutual Information (MI) score is calculated as:

    .. math::

        \\text{d(f,g)} = \\sum_{f,g} p(f,g) \\log \\frac{p(f,g)}{p(f)p(g)}

    References
    ----------
    .. [1]  Daven Vasishtan and Maya Topf, "Scoring functions for cryoEM density
            fitting", Journal of Structural Biology, vol. 174, no. 2,
            pp. 333--343, 2011. DOI: https://doi.org/10.1016/j.jsb.2011.01.012

    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def scoring_function(
        self,
        transformed_coordinates: NDArray,
        transformed_coordinates_mask: NDArray,
        in_volume: NDArray,
        in_volume_mask: NDArray,
    ) -> float:
        """
        Compute the Mutual Information score.

        Given the transformed coordinates and their associated mask, this function
        computes the mutual information between the target and template densities.

        Parameters
        ----------
        transformed_coordinates : NDArray
            Transformed coordinates.
        transformed_coordinates_mask : NDArray
            Mask for the transformed coordinates.
        in_volume : NDArray
            Binary mask indicating which ``transformed_coordinates`` are in the
            target volume.
        in_volume_mask : NDArray
            Binary mask indicating which ``transformed_coordinates`` are in the
            target mask volume.

        Returns
        -------
        float
            The Mutual Information score.
        """
        p_xy, target, template = np.histogram2d(
            self.target_density[tuple(transformed_coordinates[:, in_volume])],
            self.template_weights[in_volume],
        )
        p_x, p_y = np.sum(p_xy, axis=1), np.sum(p_xy, axis=0)

        p_xy /= p_xy.sum()
        p_x /= p_x.sum()
        p_y /= p_y.sum()

        logprob = np.divide(p_xy, p_x[:, None] * p_y[None, :] + np.finfo(float).eps)
        score = np.nansum(p_xy * logprob)

        return score


class Envelope(MatchCoordinatesToDensity):
    """
    Class representing the Envelope matching score.

    The Envelope score (ENV) is calculated as:

    .. math::

        \\text{d(f,g)} =    \\sum_{\\mathbf{p} \\in P} f'(\\mathbf{p})
                            \\cdot g'(\\mathbf{p})

    References
    ----------
    .. [1]  Daven Vasishtan and Maya Topf, "Scoring functions for cryoEM density
            fitting", Journal of Structural Biology, vol. 174, no. 2,
            pp. 333--343, 2011. DOI: https://doi.org/10.1016/j.jsb.2011.01.012
    """

    def __init__(self, target_threshold: float, **kwargs):
        super().__init__(**kwargs)
        self.target_density = np.where(self.target_density > target_threshold, -1, 1)
        self.target_density_present = np.sum(self.target_density == -1)
        self.target_density_absent = np.sum(self.target_density == 1)
        self.template_weights = np.ones_like(self.template_weights)

    def scoring_function(
        self,
        transformed_coordinates: NDArray,
        transformed_coordinates_mask: NDArray,
        in_volume: NDArray,
        in_volume_mask: NDArray,
    ) -> float:
        """
        Compute the Envelope score.

        Given the transformed coordinates and their associated mask, this function
        computes the envelope score based on target density thresholds.

        Parameters
        ----------
        transformed_coordinates : NDArray
            Transformed coordinates.
        transformed_coordinates_mask : NDArray
            Mask for the transformed coordinates.
        in_volume : NDArray
            Binary mask indicating which ``transformed_coordinates`` are in the
            target volume.
        in_volume_mask : NDArray
            Binary mask indicating which ``transformed_coordinates`` are in the
            target mask volume.

        Returns
        -------
        float
            The Envelope score.
        """
        score = self.target_density[tuple(transformed_coordinates[:, in_volume])]
        unassigned_density = self.target_density_present - (score == -1).sum()

        score = score.sum() - unassigned_density - 2 * np.sum(np.invert(in_volume))
        min_score = -self.target_density_present - 2 * self.target_density_absent
        score = (score - 2 * min_score) / (2 * self.target_density_present - min_score)

        return score


class NormalVectorScore(MatchCoordinatesToCoordinates):
    """
    Class representing the Normal Vector matching score.

    The Normal Vector Score (NVS) is calculated as:

    .. math::

        \\text{d(f,g)} = \\frac{1}{N} \\sum_{i=1}^{N}
        \\frac{
            {\\vec{f}_i} \\cdot {\\vec{g}_i}
        }{
            ||\\vec{f}_i|| \\, ||\\vec{g}_i||
        }

    References
    ----------
    .. [1]  Daven Vasishtan and Maya Topf, "Scoring functions for cryoEM density
            fitting", Journal of Structural Biology, vol. 174, no. 2,
            pp. 333--343, 2011. DOI: https://doi.org/10.1016/j.jsb.2011.01.012

    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def scoring_function(
        self,
        transformed_coordinates: NDArray,
        transformed_coordinates_mask: NDArray,
        **kwargs,
    ) -> float:
        """
        Compute the Normal Vector Score.

        Given the template and target vectors, this function computes the average
        cosine similarity between the two sets of vectors.

        Parameters
        ----------
        template_vectors : NDArray
            Normal vectors derived from the template.
        target_vectors : NDArray
            Normal vectors derived from the target.

        Returns
        -------
        float
            The Normal Vector Score.
        """
        numerator = np.multiply(transformed_coordinates, self.target_coordinates)
        denominator = np.linalg.norm(transformed_coordinates)
        denominator *= np.linalg.norm(self.target_coordinates)
        score = np.mean(numerator / denominator)
        return score


MATCHING_OPTIMIZATION_REGISTER = {
    "CrossCorrelation": CrossCorrelation,
    "LaplaceCrossCorrelation": LaplaceCrossCorrelation,
    "NormalizedCrossCorrelationMean": NormalizedCrossCorrelationMean,
    "NormalizedCrossCorrelation": NormalizedCrossCorrelation,
    "MaskedCrossCorrelation": MaskedCrossCorrelation,
    "PartialLeastSquareDifference": PartialLeastSquareDifference,
    "Envelope": Envelope,
    "Chamfer": Chamfer,
    "MutualInformation": MutualInformation,
    "NormalVectorScore": NormalVectorScore,
}


def register_matching_optimization(match_name: str, match_class: type):
    """
    Registers a class to be used by :py:class:`FitRefinement`.

    Parameters
    ----------
    match_name : str
        Name of the matching instance.
    match_class : type
        Class pointer.

    Raises
    ------
    ValueError
        If any of the required methods is not defined.
    """
    methods_to_check = ["__init__", "__call__", "scoring_function"]

    for method in methods_to_check:
        if not hasattr(match_class, method):
            raise ValueError(
                f"Method '{method}' is not defined in the provided class or object."
            )
    MATCHING_OPTIMIZATION_REGISTER[match_name] = match_class


class FitRefinement:
    """
    A class to refine the fit between target and template coordinates.

    Notes
    -----
    By default scipy.optimize.differential_evolution or scipy.optimize.basinhopping
    are used which can be unreliable if the initial alignment is very poor. Other
    optimizers can be implemented by subclassing :py:class:`FitRefinement` and
    overwriting the :py:meth:`FitRefinement.refine` function.

    """

    @staticmethod
    def map_coordinates_to_array(
        coordinates: NDArray,
        array_shape: NDArray,
        array_origin: NDArray,
        sampling_rate: NDArray,
        coordinates_mask: NDArray = None,
    ) -> Tuple[NDArray, NDArray]:
        """
        Map coordinates to a volume based on given voxel size and origin.

        Parameters
        ----------
        coordinates : NDArray
            An array representing the coordinates to be mapped [d x N].
        array_shape : NDArray
            The shape of the array to which the coordinates are mapped.
        array_origin : NDArray
            The origin of the array to which the coordinates are mapped.
        sampling_rate : NDArray
            The size of a voxel in the array.
        coordinates_mask : NDArray, optional
            An array representing the mask for the coordinates [d x T].

        Returns
        -------
        tuple
            Returns transformed coordinates, transformed coordinates mask,
            mask for in_volume points, and mask for in_volume points in mask.
        """
        coordinates = coordinates.astype(sampling_rate.dtype)
        np.divide(
            coordinates - array_origin[:, None], sampling_rate[:, None], out=coordinates
        )
        transformed_coordinates = coordinates.astype(int)
        in_volume = np.logical_and(
            transformed_coordinates < np.array(array_shape)[:, None],
            transformed_coordinates >= 0,
        ).min(axis=0)

        transformed_coordinates_mask, in_volume_mask = None, None

        if coordinates_mask is not None:
            coordinates_mask = coordinates_mask.astype(sampling_rate.dtype)
            np.divide(
                coordinates_mask - array_origin[:, None],
                sampling_rate[:, None],
                out=coordinates_mask,
            )
            transformed_coordinates_mask = coordinates_mask.astype(int)
            in_volume_mask = np.logical_and(
                transformed_coordinates_mask < np.array(array_shape)[:, None],
                transformed_coordinates_mask >= 0,
            ).min(axis=0)

        return (
            transformed_coordinates,
            transformed_coordinates_mask,
            in_volume,
            in_volume_mask,
        )

    @staticmethod
    def array_from_coordinates(
        coordinates: NDArray,
        weights: NDArray,
        sampling_rate: NDArray,
        origin: NDArray = None,
        shape: NDArray = None,
    ) -> Tuple[NDArray, NDArray, NDArray]:
        """
        Create a volume from coordinates, using given weights and voxel size.

        Parameters
        ----------
        coordinates : NDArray
            An array representing the coordinates [d x N].
        weights : NDArray
            An array representing the weights for each coordinate [N].
        sampling_rate : NDArray
            The size of a voxel in the volume.
        origin : NDArray, optional
            The origin of the volume.
        shape : NDArray, optional
            The shape of the volume.

        Returns
        -------
        tuple
            Returns the generated volume, positions of coordinates, and origin.
        """
        if origin is None:
            origin = coordinates.min(axis=1)

        positions = np.divide(coordinates - origin[:, None], sampling_rate[:, None])
        positions = positions.astype(int)

        if shape is None:
            shape = positions.max(axis=1) + 1

        arr = np.zeros(shape, dtype=np.float32)
        np.add.at(arr, tuple(positions), weights)
        return arr, positions, origin

    def refine(
        self,
        target_coordinates: NDArray,
        target_weights: NDArray,
        template_coordinates: NDArray,
        template_weights: NDArray,
        sampling_rate: float = None,
        translational_uncertainty: Tuple[float] = None,
        rotational_uncertainty: Tuple[float] = None,
        scoring_class: str = "CrossCorrelation",
        scoring_class_parameters: Dict = dict(),
        local_optimization: bool = True,
        maxiter: int = 100,
    ) -> (NDArray, NDArray):
        """
        Refines the alignment of template coordinates to target coordinates.

        Parameters
        ----------
        target_coordinates : NDArray
            The coordinates of the target.

        target_weights : NDArray
            The weights of the target.

        template_coordinates : NDArray
            The coordinates of the template.

        template_weights : NDArray
            The weights of the template.

        sampling_rate : float, optional
            The size of the voxel. Default is None.

        translational_uncertainty : (float,), optional
            The translational uncertainty. Default is None.

        rotational_uncertainty : (float,), optional
            The rotational uncertainty. Default is None.

        scoring_class : str, optional
            The scoring class to be used. Default is "CC".

        scoring_class_parameters : dict, optional
            The parameters for the scoring class. Default is an empty dictionary.

        local_optimization : bool, optional
            Whether to use local optimization. Default is True.

        maxiter : int, optional
            The maximum number of iterations. Default is 100.

        Returns
        -------
        tuple
            A tuple containing the translation and rotation matrix of the refinement,
            as well as the score of the refinement.

        Raises
        ------
        NotNotImplementedError
            If scoring class is not a part of `MATCHING_OPTIMIZATION_REGISTER`.
            Individual scores can be added via
            :py:meth:`register_matching_optimization`.

        See Also
        --------
        :py:meth:`register_matching_optimization`
        """
        if scoring_class not in MATCHING_OPTIMIZATION_REGISTER:
            raise NotImplementedError(
                f"Parameter score has to be one of "
                f"{', '.join(MATCHING_OPTIMIZATION_REGISTER.keys())}."
            )
        scoring_class = MATCHING_OPTIMIZATION_REGISTER.get(scoring_class, None)

        if sampling_rate is None:
            sampling_rate = np.ones(1)
        sampling_rate = np.repeat(
            sampling_rate, target_coordinates.shape[0] // sampling_rate.size
        )

        score = scoring_class(
            target_coordinates=target_coordinates,
            template_coordinates=template_coordinates,
            target_weights=target_weights,
            template_weights=template_weights,
            sampling_rate=sampling_rate,
            **scoring_class_parameters,
        )

        initial_score = score(np.zeros(6))

        mass_center_target = np.dot(target_coordinates, target_weights)
        mass_center_target /= target_weights.sum()
        mass_center_template = np.dot(template_coordinates, template_weights)
        mass_center_template /= template_weights.sum()

        if translational_uncertainty is None:
            mass_center_difference = np.ceil(
                np.subtract(mass_center_target, mass_center_template)
            ).astype(int)
            target_range = np.ceil(
                np.divide(
                    np.subtract(
                        target_coordinates.max(axis=1), target_coordinates.min(axis=1)
                    ),
                    2,
                )
            ).astype(int)
            translational_uncertainty = tuple(
                (center - start, center + start)
                for center, start in zip(mass_center_difference, target_range)
            )
        if rotational_uncertainty is None:
            rotational_uncertainty = tuple(
                (-90, 90) for _ in range(target_coordinates.shape[0])
            )

        uncertainty = (*translational_uncertainty, *rotational_uncertainty)
        bounds = [bound if bound != (0, 0) else (-1e-9, 1e-9) for bound in uncertainty]
        linear_constraint = LinearConstraint(
            np.eye(len(bounds)), np.min(bounds, axis=1), np.max(bounds, axis=1)
        )

        if local_optimization:
            result = basinhopping(
                x0=np.zeros(6),
                func=score,
                niter=maxiter,
                minimizer_kwargs={"method": "COBYLA", "constraints": linear_constraint},
            )
        else:
            result = differential_evolution(
                func=score,
                bounds=bounds,
                constraints=linear_constraint,
                maxiter=maxiter,
            )

        print(f"Initial score: {-initial_score} - Refined score: {-result.fun}")
        if initial_score < result.fun:
            result.x = np.zeros_like(result.x)
        translation, rotation = result.x[:3], result.x[3:]
        rotation_matrix = euler_to_rotationmatrix(rotation)
        return translation, rotation_matrix, -result.fun



from typing import Union, TypeVar

NDArray = TypeVar("numpy.ndarray")
CupyArray = TypeVar("cupy.ndarray")
TorchTensor = TypeVar("torch.Tensor")

Scalar = Union[int, float, complex]

ArrayLike = TypeVar("array_like")
CallbackClass = TypeVar("callback_class")



""" Backend using numpy and pyFFTW for template matching.

    Copyright (c) 2023 European Molecular Biology Laboratory

    Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

from typing import Tuple, Dict, List
from multiprocessing import shared_memory
from multiprocessing.managers import SharedMemoryManager
from contextlib import contextmanager

import numpy as np
from psutil import virtual_memory
from numpy.typing import NDArray
from pyfftw import zeros_aligned, simd_alignment, FFTW, next_fast_len
from pyfftw.builders import rfftn as rfftn_builder, irfftn as irfftn_builder
from scipy.ndimage import maximum_filter, affine_transform

from .matching_backend import MatchingBackend
from ..matching_utils import rigid_transform


class NumpyFFTWBackend(MatchingBackend):
    """
    A numpy and pyfftw based backend for template matching.
    """

    def __init__(
        self,
        array_backend=np,
        default_dtype=np.float32,
        complex_dtype=np.complex64,
        default_dtype_int=np.int32,
        **kwargs,
    ):
        super().__init__(
            array_backend=array_backend,
            default_dtype=default_dtype,
            complex_dtype=complex_dtype,
            default_dtype_int=np.int32,
        )
        self.affine_transform = affine_transform

    def to_backend_array(self, arr: NDArray) -> NDArray:
        if isinstance(arr, self._array_backend.ndarray):
            return arr
        return self._array_backend.asarray(arr)

    def to_numpy_array(self, arr: NDArray) -> NDArray:
        return arr

    def to_cpu_array(self, arr: NDArray) -> NDArray:
        return arr

    def free_cache(self):
        pass

    def add(self, x1, x2, *args, **kwargs) -> NDArray:
        x1 = self.to_backend_array(x1)
        x2 = self.to_backend_array(x2)
        return self._array_backend.add(x1, x2, *args, **kwargs)

    def subtract(self, x1, x2, *args, **kwargs) -> NDArray:
        x1 = self.to_backend_array(x1)
        x2 = self.to_backend_array(x2)
        return self._array_backend.subtract(x1, x2, *args, **kwargs)

    def multiply(self, x1, x2, *args, **kwargs) -> NDArray:
        x1 = self.to_backend_array(x1)
        x2 = self.to_backend_array(x2)
        return self._array_backend.multiply(x1, x2, *args, **kwargs)

    def divide(self, x1, x2, *args, **kwargs) -> NDArray:
        x1 = self.to_backend_array(x1)
        x2 = self.to_backend_array(x2)
        return self._array_backend.divide(x1, x2, *args, **kwargs)

    def mod(self, x1, x2, *args, **kwargs):
        x1 = self.to_backend_array(x1)
        x2 = self.to_backend_array(x2)
        return self._array_backend.mod(x1, x2, *args, **kwargs)

    def sum(self, *args, **kwargs) -> NDArray:
        return self._array_backend.sum(*args, **kwargs)

    def einsum(self, *args, **kwargs) -> NDArray:
        return self._array_backend.einsum(*args, **kwargs)

    def mean(self, *args, **kwargs) -> NDArray:
        return self._array_backend.mean(*args, **kwargs)

    def std(self, *args, **kwargs) -> NDArray:
        return self._array_backend.std(*args, **kwargs)

    def max(self, *args, **kwargs) -> NDArray:
        return self._array_backend.max(*args, **kwargs)

    def min(self, *args, **kwargs) -> NDArray:
        return self._array_backend.min(*args, **kwargs)

    def maximum(self, x1, x2, *args, **kwargs) -> NDArray:
        x1 = self.to_backend_array(x1)
        x2 = self.to_backend_array(x2)
        return self._array_backend.maximum(x1, x2, *args, **kwargs)

    def minimum(self, x1, x2, *args, **kwargs) -> NDArray:
        x1 = self.to_backend_array(x1)
        x2 = self.to_backend_array(x2)
        return self._array_backend.minimum(x1, x2, *args, **kwargs)

    def sqrt(self, *args, **kwargs) -> NDArray:
        return self._array_backend.sqrt(*args, **kwargs)

    def square(self, *args, **kwargs) -> NDArray:
        return self._array_backend.square(*args, **kwargs)

    def abs(self, *args, **kwargs) -> NDArray:
        return self._array_backend.abs(*args, **kwargs)

    def transpose(self, arr):
        return arr.T

    def power(self, *args, **kwargs):
        return self._array_backend.power(*args, **kwargs)

    def tobytes(self, arr):
        return arr.tobytes()

    def size(self, arr):
        return arr.size

    def fill(self, arr: NDArray, value: float) -> None:
        arr.fill(value)

    def zeros(self, shape, dtype=np.float64) -> NDArray:
        return self._array_backend.zeros(shape=shape, dtype=dtype)

    def full(self, shape, fill_value, dtype=None, **kwargs) -> NDArray:
        return self._array_backend.full(
            shape, dtype=dtype, fill_value=fill_value, **kwargs
        )

    def eps(self, dtype: type) -> NDArray:
        """
        Returns the eps defined as diffeerence between 1.0 and the next
        representable floating point value larger than 1.0.

        Parameters
        ----------
        dtype : type
            Data type for which eps should be returned.

        Returns
        -------
        Scalar
            The eps for the given data type
        """
        return self._array_backend.finfo(dtype).eps

    def datatype_bytes(self, dtype: type) -> NDArray:
        """
        Return the number of bytes occupied by a given datatype.

        Parameters
        ----------
        dtype : type
            Datatype for which the number of bytes is to be determined.

        Returns
        -------
        int
            Number of bytes occupied by the datatype.
        """
        temp = self._array_backend.zeros(1, dtype=dtype)
        return temp.nbytes

    def clip(self, *args, **kwargs) -> NDArray:
        return self._array_backend.clip(*args, **kwargs)

    def flip(self, a, axis, **kwargs):
        return self._array_backend.flip(a, axis, **kwargs)

    @staticmethod
    def astype(arr, dtype):
        return arr.astype(dtype)

    def arange(self, *args, **kwargs):
        return self._array_backend.arange(*args, **kwargs)

    def stack(self, *args, **kwargs):
        return self._array_backend.stack(*args, **kwargs)

    def concatenate(self, *args, **kwargs):
        return self._array_backend.concatenate(*args, **kwargs)

    def repeat(self, *args, **kwargs):
        return self._array_backend.repeat(*args, **kwargs)

    def topk_indices(self, arr: NDArray, k: int):
        temp = arr.reshape(-1)
        indices = self._array_backend.argpartition(temp, -k)[-k:][:k]
        sorted_indices = indices[self._array_backend.argsort(temp[indices])][::-1]
        sorted_indices = self.unravel_index(indices=sorted_indices, shape=arr.shape)
        return sorted_indices

    def indices(self, *args, **kwargs) -> NDArray:
        return self._array_backend.indices(*args, **kwargs)

    def roll(self, a, shift, axis, **kwargs):
        return self._array_backend.roll(
            a,
            shift=shift,
            axis=axis,
            **kwargs,
        )

    def unique(self, *args, **kwargs):
        return self._array_backend.unique(*args, **kwargs)

    def argsort(self, *args, **kwargs):
        return self._array_backend.argsort(*args, **kwargs)

    def unravel_index(self, indices, shape):
        return self._array_backend.unravel_index(indices=indices, shape=shape)

    def tril_indices(self, *args, **kwargs):
        return self._array_backend.tril_indices(*args, **kwargs)

    def max_filter_coordinates(self, score_space, min_distance: Tuple[int]):
        score_box = tuple(min_distance for _ in range(score_space.ndim))
        max_filter = maximum_filter(score_space, size=score_box, mode="constant")
        max_filter = max_filter == score_space

        peaks = np.array(np.nonzero(max_filter)).T
        return peaks

    @staticmethod
    def preallocate_array(shape: Tuple[int], dtype: type) -> NDArray:
        """
        Returns a byte-aligned array of zeros with specified shape and dtype.

        Parameters
        ----------
        shape : Tuple[int]
            Desired shape for the array.
        dtype : type
            Desired data type for the array.

        Returns
        -------
        NDArray
            Byte-aligned array of zeros with specified shape and dtype.
        """
        arr = zeros_aligned(shape, dtype=dtype, n=simd_alignment)
        return arr

    def sharedarr_to_arr(
        self, shape: Tuple[int], dtype: str, shm: shared_memory.SharedMemory
    ) -> NDArray:
        """
        Returns an array of given shape and dtype from shared memory location.

        Parameters
        ----------
        shape : tuple
            Tuple of integers specifying the shape of the array.
        dtype : str
            String specifying the dtype of the array.
        shm : shared_memory.SharedMemory
            Shared memory object where the array is stored.

        Returns
        -------
        NDArray
            Array of the specified shape and dtype from the shared memory location.
        """
        return self.ndarray(shape, dtype, shm.buf)

    def arr_to_sharedarr(
        self, arr: NDArray, shared_memory_handler: type = None
    ) -> shared_memory.SharedMemory:
        """
        Converts a numpy array to an object shared in memory.

        Parameters
        ----------
        arr : NDArray
            Numpy array to convert.
        shared_memory_handler : type, optional
            The type of shared memory handler. Default is None.

        Returns
        -------
        shared_memory.SharedMemory
            The shared memory object containing the numpy array.
        """
        if type(shared_memory_handler) == SharedMemoryManager:
            shm = shared_memory_handler.SharedMemory(size=arr.nbytes)
        else:
            shm = shared_memory.SharedMemory(create=True, size=arr.nbytes)
        np_array = self.ndarray(arr.shape, dtype=arr.dtype, buffer=shm.buf)
        np_array[:] = arr[:].copy()
        return shm

    def topleft_pad(self, arr: NDArray, shape: Tuple[int], padval: int = 0) -> NDArray:
        """
        Returns an array that has been padded to a specified shape with a padding
        value at the top-left corner.

        Parameters
        ----------
        arr : NDArray
            Input array to be padded.
        shape : Tuple[int]
            Desired shape for the output array.
        padval : int, optional
            Value to use for padding, default is 0.

        Returns
        -------
        NDArray
            Array that has been padded to the specified shape.
        """
        b = self.preallocate_array(shape, arr.dtype)
        self.add(b, padval, out=b)
        aind = [slice(None, None)] * arr.ndim
        bind = [slice(None, None)] * arr.ndim
        for i in range(arr.ndim):
            if arr.shape[i] > shape[i]:
                aind[i] = slice(0, shape[i])
            elif arr.shape[i] < shape[i]:
                bind[i] = slice(0, arr.shape[i])
        b[tuple(bind)] = arr[tuple(aind)]
        return b

    def build_fft(
        self,
        fast_shape: Tuple[int],
        fast_ft_shape: Tuple[int],
        real_dtype: type,
        complex_dtype: type,
        fftargs: Dict = {},
        temp_real: NDArray = None,
        temp_fft: NDArray = None,
    ) -> Tuple[FFTW, FFTW]:
        """
        Build pyFFTW builder functions.

        Parameters
        ----------
        fast_shape : tuple
            Tuple of integers corresponding to fast convolution shape
            (see `compute_convolution_shapes`).
        fast_ft_shape : tuple
            Tuple of integers corresponding to the shape of the fourier
            transform array (see `compute_convolution_shapes`).
        real_dtype : dtype
            Numpy dtype of the inverse fourier transform.
        complex_dtype : dtype
            Numpy dtype of the fourier transform.
        fftargs : dict, optional
            Dictionary passed to pyFFTW builders.
        temp_real : NDArray, optional
            Temporary real numpy array, by default None.
        temp_fft : NDArray, optional
            Temporary fft numpy array, by default None.

        Returns
        -------
        tuple
            Tuple containing callable pyFFTW objects for forward and inverse
            fourier transform.
        """
        if temp_real is None:
            temp_real = self.preallocate_array(fast_shape, real_dtype)
        if temp_fft is None:
            temp_fft = self.preallocate_array(fast_ft_shape, complex_dtype)

        default_values = {
            "planner_effort": "FFTW_MEASURE",
            "auto_align_input": False,
            "auto_contiguous": False,
            "avoid_copy": True,
            "overwrite_input": True,
            "threads": 1,
        }
        for key in default_values:
            if key in fftargs:
                continue
            fftargs[key] = default_values[key]

        rfftn = rfftn_builder(temp_real, s=fast_shape, **fftargs)

        overwrite_input = None
        if "overwrite_input" in fftargs:
            overwrite_input = fftargs.pop("overwrite_input")
        irfftn = irfftn_builder(temp_fft, s=fast_shape, **fftargs)

        if overwrite_input is not None:
            fftargs["overwrite_input"] = overwrite_input
        return rfftn, irfftn

    def extract_center(self, arr: NDArray, newshape: Tuple[int]) -> NDArray:
        """
        Extract the centered portion of an array based on a new shape.

        Parameters
        ----------
        arr : NDArray
            Input array.
        newshape : tuple
            Desired shape for the central portion.

        Returns
        -------
        NDArray
            Central portion of the array with shape `newshape`.

        References
        ----------
        .. [1] https://github.com/scipy/scipy/blob/v1.11.2/scipy/signal/_signaltools.py
        """
        new_shape = self.to_backend_array(newshape)
        current_shape = self.to_backend_array(arr.shape)
        starts = self.subtract(current_shape, new_shape)
        starts = self.astype(self.divide(starts, 2), int)
        stops = self.add(starts, newshape)
        box = tuple(slice(start, stop) for start, stop in zip(starts, stops))
        return arr[box]

    def compute_convolution_shapes(
        self, arr1_shape: Tuple[int], arr2_shape: Tuple[int]
    ) -> Tuple[List[int], List[int], List[int]]:
        """
        Computes regular, optimized and fourier convolution shape.

        Parameters
        ----------
        arr1_shape : tuple
            Tuple of integers corresponding to array1 shape.
        arr2_shape : tuple
            Tuple of integers corresponding to array2 shape.

        Returns
        -------
        tuple
            Tuple with regular convolution shape, convolution shape optimized for faster
            fourier transform, shape of the forward fourier transform
            (see :py:meth:`build_fft`).
        """
        convolution_shape = [
            int(x) + int(y) - 1 for x, y in zip(arr1_shape, arr2_shape)
        ]
        fast_shape = [next_fast_len(x) for x in convolution_shape]
        fast_ft_shape = list(fast_shape[:-1]) + [fast_shape[-1] // 2 + 1]

        return convolution_shape, fast_shape, fast_ft_shape

    def rotate_array(
        self,
        arr: NDArray,
        rotation_matrix: NDArray,
        arr_mask: NDArray = None,
        translation: NDArray = None,
        use_geometric_center: bool = False,
        out: NDArray = None,
        out_mask: NDArray = None,
        order: int = 3,
    ) -> None:
        """
        Rotates coordinates of arr according to rotation_matrix.

        If no output array is provided, this method will compute an array with
        sufficient space to hold all elements. If both `arr` and `arr_mask`
        are provided, `arr_mask` will be centered according to arr.

        Parameters
        ----------
        arr : NDArray
            The input array to be rotated.
        arr_mask : NDArray, optional
            The mask of `arr` that will be equivalently rotated.
        rotation_matrix : NDArray
            The rotation matrix to apply [d x d].
        translation : NDArray
            The translation to apply [d].
        use_geometric_center : bool, optional
            Whether the rotation should be centered around the geometric
            or mass center. Default is mass center.
        out : NDArray, optional
            The output array to write the rotation of `arr` to.
        out_mask : NDArray, optional
            The output array to write the rotation of `arr_mask` to.
        order : int, optional
            Spline interpolation order. Has to be in the range 0-5.
        """

        if order is None:
            mask_coordinates = None
            if arr_mask is not None:
                mask_coordinates = np.array(np.where(arr_mask > 0))
            return self.rotate_array_coordinates(
                arr=arr,
                arr_mask=arr_mask,
                coordinates=np.array(np.where(arr > 0)),
                mask_coordinates=mask_coordinates,
                out=out,
                out_mask=out_mask,
                rotation_matrix=rotation_matrix,
                translation=translation,
                use_geometric_center=use_geometric_center,
            )

        rotate_mask = arr_mask is not None
        return_type = (out is None) + 2 * rotate_mask * (out_mask is None)
        translation = np.zeros(arr.ndim) if translation is None else translation

        center = np.divide(arr.shape, 2)
        if not use_geometric_center:
            center = self.center_of_mass(arr, cutoff=0)

        rotation_matrix_inverted = np.linalg.inv(rotation_matrix)
        transformed_center = rotation_matrix_inverted @ center.reshape(-1, 1)
        transformed_center = transformed_center.reshape(-1)
        base_offset = np.subtract(center, transformed_center)
        offset = np.subtract(base_offset, translation)

        out = np.zeros_like(arr) if out is None else out
        out_slice = tuple(slice(0, stop) for stop in arr.shape)

        # Applying the prefilter can cause artifacts in the mask
        affine_transform(
            input=arr,
            matrix=rotation_matrix_inverted,
            offset=offset,
            mode="constant",
            output=out[out_slice],
            order=order,
            prefilter=True,
        )

        if rotate_mask:
            out_mask = np.zeros_like(arr_mask) if out_mask is None else out_mask
            out_mask_slice = tuple(slice(0, stop) for stop in arr_mask.shape)
            affine_transform(
                input=arr_mask,
                matrix=rotation_matrix_inverted,
                offset=offset,
                mode="constant",
                output=out_mask[out_mask_slice],
                order=order,
                prefilter=False,
            )

        match return_type:
            case 0:
                return None
            case 1:
                return out
            case 2:
                return out_mask
            case 3:
                return out, out_mask

    @staticmethod
    def rotate_array_coordinates(
        arr: NDArray,
        coordinates: NDArray,
        rotation_matrix: NDArray,
        translation: NDArray = None,
        out: NDArray = None,
        use_geometric_center: bool = True,
        arr_mask: NDArray = None,
        mask_coordinates: NDArray = None,
        out_mask: NDArray = None,
    ) -> None:
        """
        Rotates coordinates of arr according to rotation_matrix.

        If no output array is provided, this method will compute an array with
        sufficient space to hold all elements. If both `arr` and `arr_mask`
        are provided, `arr_mask` will be centered according to arr.

        No centering will be performed if the rotation matrix is the identity matrix.

        Parameters
        ----------
        arr : NDArray
            The input array to be rotated.
        coordinates : NDArray
            The pointcloud [d x N] containing elements of `arr` that should be rotated.
            See :py:meth:`Density.to_pointcloud` on how to obtain the coordinates.
        rotation_matrix : NDArray
            The rotation matrix to apply [d x d].
        rotation_matrix : NDArray
            The translation to apply [d].
        out : NDArray, optional
            The output array to write the rotation of `arr` to.
        use_geometric_center : bool, optional
            Whether the rotation should be centered around the geometric
            or mass center.
        arr_mask : NDArray, optional
            The mask of `arr` that will be equivalently rotated.
        mask_coordinates : NDArray, optional
            Equivalent to `coordinates`, but containing elements of `arr_mask`
            that should be rotated.
        out_mask : NDArray, optional
            The output array to write the rotation of `arr_mask` to.
        """
        rotate_mask = arr_mask is not None and mask_coordinates is not None
        return_type = (out is None) + 2 * rotate_mask * (out_mask is None)

        # Otherwise array might be slightly shifted by centering
        if np.allclose(
            rotation_matrix,
            np.eye(rotation_matrix.shape[0], dtype=rotation_matrix.dtype),
        ):
            center_rotation = False

        coordinates_rotated = np.empty(coordinates.shape, dtype=rotation_matrix.dtype)
        mask_rotated = (
            np.empty(mask_coordinates.shape, dtype=rotation_matrix.dtype)
            if rotate_mask
            else None
        )

        center = np.array(arr.shape) // 2 if use_geometric_center else None
        if translation is None:
            translation = np.zeros(coordinates_rotated.shape[0])

        rigid_transform(
            coordinates=coordinates,
            coordinates_mask=mask_coordinates,
            out=coordinates_rotated,
            out_mask=mask_rotated,
            rotation_matrix=rotation_matrix,
            translation=translation,
            use_geometric_center=use_geometric_center,
            center=center,
        )

        coordinates_rotated = coordinates_rotated.astype(int)
        offset = coordinates_rotated.min(axis=1)
        np.multiply(offset, offset < 0, out=offset)
        coordinates_rotated -= offset[:, None]

        out_offset = np.zeros(
            coordinates_rotated.shape[0], dtype=coordinates_rotated.dtype
        )
        if out is None:
            out_offset = coordinates_rotated.min(axis=1)
            coordinates_rotated -= out_offset[:, None]
            out = np.zeros(coordinates_rotated.max(axis=1) + 1, dtype=arr.dtype)

        if rotate_mask:
            mask_rotated = mask_rotated.astype(int)
            if out_mask is None:
                mask_rotated -= out_offset[:, None]
                out_mask = np.zeros(
                    coordinates_rotated.max(axis=1) + 1, dtype=arr.dtype
                )

            in_box = np.logical_and(
                mask_rotated < np.array(out_mask.shape)[:, None],
                mask_rotated >= 0,
            ).min(axis=0)
            out_of_box = np.invert(in_box).sum()
            if out_of_box != 0:
                print(
                    f"{out_of_box} elements out of bounds. Perhaps increase"
                    " *arr_mask* size."
                )

            mask_coordinates = tuple(mask_coordinates[:, in_box])
            mask_rotated = tuple(mask_rotated[:, in_box])
            np.add.at(out_mask, mask_rotated, arr_mask[mask_coordinates])

        # Negative coordinates would be (mis)interpreted as reverse index
        in_box = np.logical_and(
            coordinates_rotated < np.array(out.shape)[:, None], coordinates_rotated >= 0
        ).min(axis=0)
        out_of_box = np.invert(in_box).sum()
        if out_of_box != 0:
            print(f"{out_of_box} elements out of bounds. Perhaps increase *out* size.")

        coordinates = coordinates[:, in_box]
        coordinates_rotated = coordinates_rotated[:, in_box]

        coordinates = tuple(coordinates)
        coordinates_rotated = tuple(coordinates_rotated)
        np.add.at(out, coordinates_rotated, arr[coordinates])

        match return_type:
            case 0:
                return None
            case 1:
                return out
            case 2:
                return out_mask
            case 3:
                return out, out_mask

    def center_of_mass(self, arr: NDArray, cutoff: float = None) -> NDArray:
        """
        Computes the center of mass of a numpy ndarray instance using all available
        elements. For template matching it typically makes sense to only input
        positive densities.

        Parameters
        ----------
        arr : NDArray
            Array to compute the center of mass of.
        cutoff : float, optional
            Densities less than or equal to cutoff are nullified for center
            of mass computation. By default considers all values.

        Returns
        -------
        NDArray
            Center of mass with shape (arr.ndim).
        """
        cutoff = arr.min() - 1 if cutoff is None else cutoff
        arr = self._array_backend.where(arr > cutoff, arr, 0)
        denominator = self.sum(arr)
        grids = self._array_backend.ogrid[tuple(slice(0, i) for i in arr.shape)]
        grids = [grid.astype(self._default_dtype) for grid in grids]

        center_of_mass = self.array(
            [
                self.sum(self.multiply(arr, grids[dim])) / denominator
                for dim in range(arr.ndim)
            ]
        )

        return center_of_mass

    def get_available_memory(self) -> int:
        return virtual_memory().available

    @contextmanager
    def set_device(self, device_index: int):
        yield None

    def device_count(self) -> int:
        return 1

    @staticmethod
    def reverse(arr: NDArray) -> NDArray:
        """
        Reverse the order of elements in an array along all its axes.

        Parameters
        ----------
        arr : NDArray
            Input array.

        Returns
        -------
        NDArray
            Reversed array.
        """
        return arr[(slice(None, None, -1),) * arr.ndim]



""" Compute memory consumption of template matching components.

    Copyright (c) 2023 European Molecular Biology Laboratory

    Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

from abc import ABC, abstractmethod
from typing import Tuple

import numpy as np
from pyfftw import next_fast_len


class MatchingMemoryUsage(ABC):
    """
    Base class for estimating the memory usage of template matching.

    This class provides a template for estimating memory usage for
    different matching methods. Users should subclass it and implement the
    `base_usage` and `per_fork` methods to specify custom memory usage
    estimates.

    Parameters
    ----------
    fast_shape : tuple of int
        Shape of the real array.
    ft_shape : tuple of int
        Shape of the complex array.
    float_nbytes : int
        Number of bytes of the used float, e.g. 4 for float32.
    complex_nbytes : int
        Number of bytes of the used complex, e.g. 8 for complex64.
    integer_nbytes : int
        Number of bytes of the used integer, e.g. 4 for int32.

    Attributes
    ----------
    real_array_size : int
        Number of elements in real array.
    complex_array_size : int
        Number of elements in complex array.
    float_nbytes : int
        Number of bytes of the used float, e.g. 4 for float32.
    complex_nbytes : int
        Number of bytes of the used complex, e.g. 8 for complex64.
    integer_nbytes : int
        Number of bytes of the used integer, e.g. 4 for int32.

    Methods
    -------
    base_usage():
        Returns the base memory usage in bytes.
    per_fork():
        Returns the memory usage in bytes per fork.
    """

    def __init__(
        self,
        fast_shape: Tuple[int],
        ft_shape: Tuple[int],
        float_nbytes: int,
        complex_nbytes: int,
        integer_nbytes: int,
    ):
        self.real_array_size = np.prod(fast_shape)
        self.complex_array_size = np.prod(ft_shape)
        self.float_nbytes = float_nbytes
        self.complex_nbytes = complex_nbytes
        self.integer_nbytes = integer_nbytes

    @abstractmethod
    def base_usage(self) -> int:
        """Return the base memory usage in bytes."""

    @abstractmethod
    def per_fork(self) -> int:
        """Return the memory usage per fork in bytes."""


class CCMemoryUsage(MatchingMemoryUsage):
    """
    Memory usage estimation for the CC fitter.

    See Also
    --------
    :py:meth:`tme.matching_exhaustive.cc_setup`.
    """

    def base_usage(self) -> int:
        float_arrays = self.real_array_size * self.float_nbytes
        complex_arrays = self.complex_array_size * self.complex_nbytes
        return float_arrays + complex_arrays

    def per_fork(self) -> int:
        float_arrays = self.real_array_size * self.float_nbytes
        complex_arrays = self.complex_array_size * self.complex_nbytes
        return float_arrays + complex_arrays


class LCCMemoryUsage(CCMemoryUsage):
    """
    Memory usage estimation for LCC scoring.
    See Also
    --------
    :py:meth:`tme.matching_exhaustive.lcc_setup`.
    """


class CORRMemoryUsage(MatchingMemoryUsage):
    """
    Memory usage estimation for CORR scoring.

    See Also
    --------
    :py:meth:`tme.matching_exhaustive.corr_setup`.
    """

    def base_usage(self) -> int:
        float_arrays = self.real_array_size * self.float_nbytes * 4
        complex_arrays = self.complex_array_size * self.complex_nbytes
        return float_arrays + complex_arrays

    def per_fork(self) -> int:
        float_arrays = self.real_array_size * self.float_nbytes
        complex_arrays = self.complex_array_size * self.complex_nbytes
        return float_arrays + complex_arrays


class CAMMemoryUsage(CORRMemoryUsage):
    """
    Memory usage estimation for CAM scoring.

    See Also
    --------
    :py:meth:`tme.matching_exhaustive.cam_setup`.
    """


class FLCSphericalMaskMemoryUsage(CORRMemoryUsage):
    """
    Memory usage estimation for FLCMSphericalMask scoring.

    See Also
    --------
    :py:meth:`tme.matching_exhaustive.flcSphericalMask_setup`.
    """


class FLCMemoryUsage(MatchingMemoryUsage):
    """
    Memory usage estimation for FLC scoring.

    See Also
    --------
    :py:meth:`tme.matching_exhaustive.flc_setup`.
    """

    def base_usage(self) -> int:
        float_arrays = self.real_array_size * self.float_nbytes * 2
        complex_arrays = self.complex_array_size * self.complex_nbytes * 2
        return float_arrays + complex_arrays

    def per_fork(self) -> int:
        float_arrays = self.real_array_size * self.float_nbytes * 3
        complex_arrays = self.complex_array_size * self.complex_nbytes * 2
        return float_arrays + complex_arrays


class MCCMemoryUsage(MatchingMemoryUsage):
    """
    Memory usage estimation for MCC scoring.

    See Also
    --------
    :py:meth:`tme.matching_exhaustive.mcc_setup`.
    """

    def base_usage(self) -> int:
        float_arrays = self.real_array_size * self.float_nbytes * 2
        complex_arrays = self.complex_array_size * self.complex_nbytes * 3
        return float_arrays + complex_arrays

    def per_fork(self) -> int:
        float_arrays = self.real_array_size * self.float_nbytes * 6
        complex_arrays = self.complex_array_size * self.complex_nbytes
        return float_arrays + complex_arrays


class MaxScoreOverRotationsMemoryUsage(MatchingMemoryUsage):
    """
    Memory usage estimation MaxScoreOverRotations Analyzer.

    See Also
    --------
    :py:class:`tme.analyzer.MaxScoreOverRotations`.
    """

    def base_usage(self) -> int:
        float_arrays = self.real_array_size * self.float_nbytes * 2
        return float_arrays

    def per_fork(self) -> int:
        return 0


class PeakCallerMaximumFilterMemoryUsage(MatchingMemoryUsage):
    """
    Memory usage estimation MaxScoreOverRotations Analyzer.

    See Also
    --------
    :py:class:`tme.analyzer.PeakCallerMaximumFilter`.
    """

    def base_usage(self) -> int:
        float_arrays = self.real_array_size * self.float_nbytes
        return float_arrays

    def per_fork(self) -> int:
        float_arrays = self.real_array_size * self.float_nbytes
        return float_arrays


class CupyBackendMemoryUsage(MatchingMemoryUsage):
    """
    Memory usage estimation for CupyBackend.

    See Also
    --------
    :py:class:`tme.backends.CupyBackend`.
    """

    def base_usage(self) -> int:
        # FFT plans, overhead from assigning FFT result, rotation interpolation
        complex_arrays = self.real_array_size * self.complex_nbytes * 3
        float_arrays = self.complex_array_size * self.float_nbytes * 2
        return float_arrays + complex_arrays

    def per_fork(self) -> int:
        return 0


def _compute_convolution_shapes(
    arr1_shape: Tuple[int], arr2_shape: Tuple[int]
) -> Tuple[Tuple[int], Tuple[int], Tuple[int]]:
    """
    Computes regular, optimized and fourier convolution shape.

    Parameters
    ----------
    arr1_shape : tuple
        Tuple of integers corresponding to array1 shape.
    arr2_shape : tuple
        Tuple of integers corresponding to array2 shape.

    Returns
    -------
    tuple
        Tuple with regular convolution shape, convolution shape optimized for faster
        fourier transform, shape of the forward fourier transform
        (see :py:meth:`build_fft`).
    """
    convolution_shape = np.add(arr1_shape, arr2_shape) - 1
    fast_shape = [next_fast_len(x) for x in convolution_shape]
    fast_ft_shape = list(fast_shape[:-1]) + [fast_shape[-1] // 2 + 1]

    return convolution_shape, fast_shape, fast_ft_shape


MATCHING_MEMORY_REGISTRY = {
    "CC": CCMemoryUsage,
    "LCC": LCCMemoryUsage,
    "CORR": CORRMemoryUsage,
    "CAM": CAMMemoryUsage,
    "MCC": MCCMemoryUsage,
    "FLCSphericalMask": FLCSphericalMaskMemoryUsage,
    "FLC": FLCMemoryUsage,
    "MaxScoreOverRotations": MaxScoreOverRotationsMemoryUsage,
    "PeakCallerMaximumFilter": PeakCallerMaximumFilterMemoryUsage,
    "cupy": CupyBackendMemoryUsage,
    "pytorch": CupyBackendMemoryUsage,
}


def estimate_ram_usage(
    shape1: Tuple[int],
    shape2: Tuple[int],
    matching_method: str,
    ncores: int,
    analyzer_method: str = None,
    backend: str = None,
    float_nbytes: int = 4,
    complex_nbytes: int = 8,
    integer_nbytes: int = 4,
) -> int:
    """
    Estimate the RAM usage for a given convolution operation based on input shapes,
    matching_method, and number of cores.

    Parameters
    ----------
    shape1 : tuple
        The shape of the input target.
    shape2 : tuple
        The shape of the input template.
    matching_method : str
        The method used for the operation.
    is_gpu : bool, optional
        Whether the computation is performed on GPU. This factors in FFT
        plan caching.
    analyzer_method : str, optional
        The method used for score analysis.
    backend : str, optional
        Backend used for computation.
    ncores : int
        The number of CPU cores used for the operation.
    float_nbytes : int
        Number of bytes of the used float, e.g. 4 for float32.
    complex_nbytes : int
        Number of bytes of the used complex, e.g. 8 for complex64.
    integer_nbytes : int
        Number of bytes of the used integer, e.g. 4 for int32.

    Returns
    -------
    int
        The estimated RAM usage for the operation in bytes.

    Notes
    -----
        Residual memory from other objects that may remain allocated during
        template matching, e.g. the full sized target when using splitting,
        are not considered by this function.

    Raises
    ------
    ValueError
        If an unsupported matching_methode is provided.
    """
    if matching_method not in MATCHING_MEMORY_REGISTRY:
        raise ValueError(
            f"Supported options are {','.join(MATCHING_MEMORY_REGISTRY.keys())}"
        )

    convolution_shape, fast_shape, ft_shape = _compute_convolution_shapes(
        shape1, shape2
    )

    memory_instance = MATCHING_MEMORY_REGISTRY[matching_method](
        fast_shape=fast_shape,
        ft_shape=ft_shape,
        float_nbytes=float_nbytes,
        complex_nbytes=complex_nbytes,
        integer_nbytes=integer_nbytes,
    )

    nbytes = memory_instance.base_usage() + memory_instance.per_fork() * ncores

    analyzer_instance = MATCHING_MEMORY_REGISTRY.get(analyzer_method, None)
    if analyzer_instance is not None:
        analyzer_instance = analyzer_instance(
            fast_shape=fast_shape,
            ft_shape=ft_shape,
            float_nbytes=float_nbytes,
            complex_nbytes=complex_nbytes,
            integer_nbytes=integer_nbytes,
        )
        nbytes += analyzer_instance.base_usage() + analyzer_instance.per_fork() * ncores

    backend_instance = MATCHING_MEMORY_REGISTRY.get(backend, None)
    if backend_instance is not None:
        backend_instance = backend_instance(
            fast_shape=fast_shape,
            ft_shape=ft_shape,
            float_nbytes=float_nbytes,
            complex_nbytes=complex_nbytes,
            integer_nbytes=integer_nbytes,
        )
        nbytes += backend_instance.base_usage() + backend_instance.per_fork() * ncores

    return nbytes



""" pyTME backend manager.

    Copyright (c) 2023 European Molecular Biology Laboratory

    Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

from typing import Dict, List

from .matching_backend import MatchingBackend
from .npfftw_backend import NumpyFFTWBackend
from .pytorch_backend import PytorchBackend
from .cupy_backend import CupyBackend


class BackendManager:
    """
    Manager for template matching backends.

    This class serves as an interface to various computational backends (e.g.,
    CPU, GPU). It allows users to seamlessly swap between different backend
    implementations while preserving the consistency and functionality of the API.
    Direct attribute and method calls to the manager are delegated to the current
    active backend.

    Attributes
    ----------
    _BACKEND_REGISTRY : dict
        A dictionary mapping backend names to their respective classes or instances.
    _backend : instance of MatchingBackend
        An instance of the currently active backend.
    _backend_name : str
        Name of the current backend.
    _backend_args : Dict
        Arguments passed to create current backend.

    Examples
    --------
    >>> from tme.backends import backend
    >>> backend.multiply(arr1, arr2)
    # This will use the default NumpyFFTWBackend's multiply method

    >>> backend.change_backend("pytorch")
    >>> backend.multiply(arr1, arr2)
    # This will use the GPUBackend's multiply method

    Notes
    -----
    To add custom backends, use the `add_backend` method. To switch
    between backends, use the `change_backend` method. Note that the backend
    has to be reinitialzed when using fork-based parallelism.
    """

    def __init__(self):
        self._BACKEND_REGISTRY = {
            "cpu_backend": NumpyFFTWBackend,
            "pytorch": PytorchBackend,
            "cupy": CupyBackend,
        }
        self._backend = NumpyFFTWBackend()
        self._backend_name = "cpu_backend"
        self._backend_args = {}

    def __repr__(self):
        return f"<BackendManager: using {self._backend_name}>"

    def __getattr__(self, name):
        return getattr(self._backend, name)

    def __dir__(self) -> List:
        """
        Return a list of attributes available in this object,
        including those from the backend.

        Returns
        -------
        list
            Sorted list of attributes.
        """
        base_attributes = []
        base_attributes.extend(dir(self.__class__))
        base_attributes.extend(self.__dict__.keys())
        base_attributes.extend(dir(self._backend))
        return sorted(base_attributes)

    def add_backend(self, backend_name: str, backend_class: type):
        """
        Adds a custom backend to the registry.

        Parameters
        ----------
        backend_name : str
            Name by which the backend can be referenced.
        backend_class : :py:class:`MatchingBackend`
            An instance of the backend to be added.

        Raises
        ------
        ValueError
            If the provided backend_instance does not inherit from MatchingBackend.
        """
        if not issubclass(backend_class, MatchingBackend):
            raise ValueError("backend_class needs to inherit from MatchingBackend.")
        self._BACKEND_REGISTRY[backend_name] = backend_class

    def change_backend(self, backend_name: str, **backend_kwargs: Dict) -> None:
        """
        Change the backend.

        Parameters
        ----------
        backend_name : str
            Name of the new backend that should be used.
        **backend_kwargs : Dict, optional
            Parameters passed to __init__ method of backend.

        Raises
        ------
        NotImplementedError
            If no backend is found with the provided name.
        """
        if backend_name not in self._BACKEND_REGISTRY:
            available_backends = ", ".join(
                [str(x) for x in self._BACKEND_REGISTRY.keys()]
            )
            raise NotImplementedError(
                f"Available backends are {available_backends} - not {backend_name}."
            )
        self._backend = self._BACKEND_REGISTRY[backend_name](**backend_kwargs)
        self._backend_name = backend_name
        self._backend_args = backend_kwargs


backend = BackendManager()



""" Implements Preprocessor class for filtering operations.

    Copyright (c) 2023 European Molecular Biology Laboratory

    Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

import inspect
from typing import Dict, Tuple

import numpy as np
from numpy.typing import NDArray

from scipy.ndimage import (
    fourier_gaussian,
    gaussian_filter,
    rank_filter,
    zoom,
    generic_gradient_magnitude,
    sobel,
    prewitt,
    laplace,
    gaussian_laplace,
    gaussian_gradient_magnitude,
)
from scipy.signal import convolve, decimate
from scipy.optimize import differential_evolution
from pywt import wavelist, wavedecn, waverecn
from scipy.interpolate import interp1d

from .density import Density
from .helpers import (
    window_kaiserb,
    window_blob,
    apply_window_filter,
    Ntree,
)
from .matching_utils import euler_to_rotationmatrix


class Preprocessor:
    """
    Implements filtering operations on density arrays.
    """

    def apply_method(self, method: str, parameters: Dict):
        """
        Apply a method on the atomic structure.

        Parameters
        ----------
        method : str
            The name of the method to be used.
        parameters : dict
            The parameters for the specified method.

        Returns
        -------
        The output of ``method``.

        Raises
        ------
        NotImplementedError
            If ``method`` is not a member of :py:class:`Preprocessor`.
        """
        if not hasattr(self, method):
            raise NotImplementedError(
                f"'{method}' is not supported as a filter method on this class."
            )
        method_to_call = getattr(self, method)
        return method_to_call(**parameters)

    def method_to_id(self, method: str, parameters: Dict) -> str:
        """
        Generate a unique ID for a specific method operation.

        Parameters
        ----------
        method : str
            The name of the method.
        parameters : dict
            A dictionary containing the parameters used by the method.

        Returns
        -------
        str
            A string representation of the method operation, which can be used
            as a unique identifier.

        Raises
        ------
        NotImplementedError
            If ``method`` is not a member of :py:class:`Preprocessor`.
        """
        if not hasattr(self, method):
            raise NotImplementedError(
                f"'{method}' is not supported as a filter method on this class."
            )
        signature = inspect.signature(getattr(self, method))
        default = {
            k: v.default
            for k, v in signature.parameters.items()
            if v.default is not inspect.Parameter.empty
        }

        default.update(parameters)

        return "-".join([str(default[key]) for key in sorted(default.keys())])

    @staticmethod
    def _gaussian_fourier(template: NDArray, sigma: NDArray) -> NDArray:
        """
        Apply a Gaussian filter in Fourier space on the provided template.

        Parameters
        ----------
        template : NDArray
            The input template on which to apply the filter.
        sigma : NDArray
            The standard deviation for Gaussian kernel. The greater the value,
            the more spread out is the filter.

        Returns
        -------
        NDArray
            The template after applying the Fourier Gaussian filter.
        """
        fourrier_map = fourier_gaussian(np.fft.fftn(template), sigma)
        template = np.real(np.fft.ifftn(fourrier_map))

        return template

    @staticmethod
    def _gaussian_real(
        template: NDArray, sigma: NDArray, cutoff_value: float = 4.0
    ) -> NDArray:
        """
        Apply a Gaussian filter on the provided template in real space.

        Parameters
        ----------
        template : NDArray
            The input template on which to apply the filter.
        sigma : NDArray
            The standard deviation for Gaussian kernel. The greater the value,
            the more spread out is the filter.
        cutoff_value : float, optional
            The value below which the data should be ignored. Default is 4.0.

        Returns
        -------
        NDArray
            The template after applying the Gaussian filter in real space.
        """
        template = gaussian_filter(template, sigma, cval=cutoff_value)
        return template

    def gaussian_filter(
        self,
        template: NDArray,
        sigma: NDArray,
        fourier: bool = False,
    ) -> NDArray:
        """
        Convolve an atomic structure with a Gaussian kernel.

        Parameters
        ----------
        template : NDArray
            The input atomic structure map.
        resolution : float, optional
            The resolution. The product of `resolution` and `sigma_coeff` is used
            to compute the `sigma` for the discretized Gaussian. Default is None.
        sigma : NDArray
            The standard deviation for Gaussian kernel. Should either be a scalar
            or a sequence of scalars.
        fourier : bool, optional
            If true, applies a Fourier Gaussian filter; otherwise, applies a
            real-space Gaussian filter. Default is False.

        Returns
        -------
        NDArray
            The simulated electron densities after applying the Gaussian filter.
        """
        sigma = 0 if sigma is None else sigma

        if sigma <= 0:
            return template

        func = self._gaussian_real if not fourier else self._gaussian_fourier
        template = func(template, sigma)

        return template

    def difference_of_gaussian_filter(
        self, template: NDArray, low_sigma: NDArray, high_sigma: NDArray
    ) -> NDArray:
        """
        Apply the Difference of Gaussian (DoG) bandpass filter on
        the provided template.

        Parameters
        ----------
        template : NDArray
            The input template on which to apply the technique.
        low_sigma : NDArray
            The smaller standard deviation for the Gaussian kernel.
            Should be scalar or sequence of scalars of length template.ndim.
        high_sigma : NDArray
            The larger standard deviation for the Gaussian kernel.
            Should be scalar or sequence of scalars of length template.ndim.

        Returns
        -------
        NDArray
            The result of applying the Difference of Gaussian technique on the template.
        """
        if np.any(low_sigma > high_sigma):
            print("low_sigma should be smaller than high_sigma.")
        im1 = self._gaussian_real(template, low_sigma)
        im2 = self._gaussian_real(template, high_sigma)
        return im1 - im2

    def local_gaussian_alignment_filter(
        self,
        target: NDArray,
        template: NDArray,
        lbd: float,
        sigma_range: Tuple[float, float] = (0.1, 20),
    ) -> NDArray:
        """
        Simulate electron density by optimizing a sum of Gaussians.

        For that, the following minimization problem is considered:

        .. math::
            dl_{\\text{{target}}} = \\frac{\\lambda}{\\sigma_{x}^{2}} + \\epsilon^{2}

        .. math::
            \\epsilon^{2} = \\| \\text{target} - \\text{template} \\|^{2}

        Parameters
        ----------
        target : NDArray
            The target electron density map.
        template : NDArray
            The input atomic structure map.
        lbd : float
            The lambda hyperparameter.
        sigma_range : tuple of float, optional
            The range of sigma values for the optimizer. Default is (0.1, 20).

        Returns
        -------
        NDArray
            Simulated electron densities.

        References
        ----------
        .. [1]  Gomez, G (Jan. 2000). Local Smoothness in terms of Variance:
                The Adaptive Gaussian Filter. In Procedings of the British
                Machine Vision Conference 2000.
        """

        class _optimizer(Preprocessor):
            def __init__(self, target, template, lbd):
                self._target = target
                self._template = template
                self._dl = np.full(template.shape, 10**9)
                self._filter = np.zeros_like(template)
                self._lbd = lbd

            def __call__(self, x, *args):
                x = x[0]
                filter = super().gaussian_filter(sigma=x, template=template)
                dl = self._lbd / (x**2) + np.power(self._target - filter, 2)
                ind = dl < self._dl
                self._dl[ind] = dl[ind]
                self._filter[ind] = filter[ind]
                return np.sum(self._dl)

        # This method needs pre normalization
        template = template.copy()
        target = target.copy()
        sd_target = np.std(target)
        sd_template = np.std(template)
        m_target = np.mean(target)
        m_template = np.mean(target)
        if sd_target != 0:
            target = (target - m_target) / sd_target

        if sd_template != 0:
            template = (template - m_template) / sd_template

        temp = _optimizer(target=target, template=template, lbd=lbd)

        _ = differential_evolution(temp, bounds=[sigma_range], seed=2)

        # Make sure there is no negative density
        temp._filter += np.abs(np.min(temp._filter))

        return temp._filter

    def local_gaussian_filter(
        self,
        template: NDArray,
        lbd: float,
        sigma_range: Tuple[float, float],
        gaussian_sigma: float,
    ) -> NDArray:
        """
        Wrapper for `Preprocessor.local_gaussian_alignment_filter` if no
        target is available.

        Parameters
        ----------
        template : NDArray
            The input atomic structure map.
        apix : float
            Ångstrom per voxel passed to `Preprocessor.gaussian_filter`.
        lbd : float
            The lambda hyperparameter, common values: 2, 5, 20.
        sigma_range : tuple of float
            The range of sigma values for the optimizer.
        gaussian_sigma : float
            The sigma value passed to `Preprocessor.gaussian_filter` to
            obtain a target.

        Returns
        -------
        NDArray
            Simulated electron densities.
        """
        filtered_data = self.gaussian_filter(sigma=gaussian_sigma, template=template)
        return self.local_gaussian_alignment_filter(
            target=filtered_data,
            template=template,
            lbd=lbd,
            sigma_range=sigma_range,
        )

    def edge_gaussian_filter(
        self,
        template: NDArray,
        edge_algorithm: str,
        sigma: float,
        reverse: bool = False,
    ) -> NDArray:
        """
        Perform Gaussian filterring according to edges in the input template.

        Parameters
        ----------
        template : NDArray
            The input atomic structure map.
        sigma : NDArray
            The sigma value for the Gaussian filter.
        edge_algorithm : str
            The algorithm used to identify edges.  Options are:

            +-------------------+------------------------------------------------+
            | 'sobel'           | Applies sobel filter for edge detection.       |
            +-------------------+------------------------------------------------+
            | 'prewitt'         | Applies prewitt filter for edge detection.     |
            +-------------------+------------------------------------------------+
            | 'laplace'         | Computes edges as second derivative.           |
            +-------------------+------------------------------------------------+
            | 'gaussian'        | See scipy.ndimage.gaussian_gradient_magnitude  |
            +-------------------+------------------------------------------------+
            | 'gaussian_laplace | See scipy.ndimage.gaussian_laplace             |
            +-------------------+------------------------------------------------+

        reverse : bool, optional
            If true, the filterring is strong along edges. Default is False.

        Returns
        -------
        NDArray
            Simulated electron densities.
        """
        if edge_algorithm == "sobel":
            edges = generic_gradient_magnitude(template, sobel)
        elif edge_algorithm == "prewitt":
            edges = generic_gradient_magnitude(template, prewitt)
        elif edge_algorithm == "laplace":
            edges = laplace(template)
        elif edge_algorithm == "gaussian":
            edges = gaussian_gradient_magnitude(template, sigma / 2)
        elif edge_algorithm == "gaussian_laplace":
            edges = gaussian_laplace(template, sigma / 2)
        else:
            raise ValueError(
                "Supported edge_algorithm values are"
                "'sobel', 'prewitt', 'laplace', 'gaussian_laplace', 'gaussian'"
            )
        edges[edges != 0] = 1
        edges /= edges.max()

        edges = gaussian_filter(edges, sigma)
        filter = gaussian_filter(template, sigma)

        if not reverse:
            res = template * edges + filter * (1 - edges)
        else:
            res = template * (1 - edges) + filter * (edges)

        return res

    def ntree_filter(
        self,
        template: NDArray,
        sigma_range: Tuple[float, float],
        target: NDArray = None,
    ) -> NDArray:
        """
        Use dyadic tree to identify volume partitions in *template*
        and filter them with respect to their occupancy.

        Parameters
        ----------
        template : NDArray
            The input atomic structure map.
        sigma_range : tuple of float
            Range of sigma values used to filter volume partitions.
        target : NDArray, optional
            If provided, dyadic tree is computed on target rather than template.

        Returns
        -------
        NDArray
            Simulated electron densities.
        """
        if target is None:
            target = template

        tree = Ntree(target)

        filter = tree.filter_chunks(arr=template, sigma_range=sigma_range)

        return filter

    def mean_filter(self, template: NDArray, width: NDArray) -> NDArray:
        """
        Perform mean filtering.

        Parameters
        ----------
        template : NDArray
            The input atomic structure map.
        width : NDArray
            Width of the mean filter along each axis. Can either have length
            one or template.ndim.

        Returns
        -------
        NDArray
            Simulated electron densities.
        """
        template = template.copy()
        interpolation_box = template.shape

        width = np.array(width)
        filter_width = np.repeat(width, template.ndim // width.size)
        filter_mask = np.ones(filter_width)
        filter_mask = filter_mask / np.sum(filter_mask)
        template = convolve(template, filter_mask, mode="same")

        # Sometimes scipy messes up the box sizes ...
        template = self.interpolate_box(box=interpolation_box, arr=template)

        return template

    def kaiserb_filter(self, template: NDArray, width: int) -> NDArray:
        """
        Apply Kaiser filter defined as:

        .. math::
            f_{kaiser} = \\frac{I_{0}(\\beta\\sqrt{1-
            \\frac{4n^{2}}{(M-1)^{2}}})}{I_{0}(\\beta)}
            -\\frac{M-1}{2} \\leq n \\leq \\frac{M-1}{2}
            \\text{With } \\beta=3.2

        Parameters
        ----------
        template : NDArray
            The input atomic structure map.
        width : int
            Width of the filter window.
        normalize : bool, optional
            If true, the output is z-transformed. Default is False.

        Returns
        -------
        NDArray
            Simulated electron densities.

        References
        ----------
        .. [1]  Sorzano, Carlos et al (Mar. 2015). Fast and accurate conversion
                of atomic models into electron density maps. AIMS Biophysics
                2, 8–20.
        """
        template, interpolation_box = template.copy(), template.shape

        kaiser_window = window_kaiserb(width=width)
        template = apply_window_filter(arr=template, filter_window=kaiser_window)

        if not np.all(template.shape == interpolation_box):
            template = self.interpolate_box(box=interpolation_box, arr=template)

        return template

    def blob_filter(self, template: NDArray, width: int) -> NDArray:
        """
        Apply blob filter defined as:

        .. math::
            f_{blob} = \\frac{\\sqrt{1-(\\frac{4n^{2}}{(M-1)^{2}})^{m}} I_{m}
            (\\beta\\sqrt{1-(\\frac{4n^{2}}{(M-1)^{2}})})}
            {I_{m}(\\beta)}
            -\\frac{M-1}{2} \\leq n \\leq \\frac{M-1}{2}
            \\text{With } \\beta=3.2 \\text{ and order=2}

        Parameters
        ----------
        template : NDArray
            The input atomic structure map.
        width : int
            Width of the filter window.

        Returns
        -------
        NDArray
            Simulated electron densities.

        References
        ----------
        .. [1]  Sorzano, Carlos et al (Mar. 2015). Fast and accurate conversion
                of atomic models into electron density maps. AIMS Biophysics
                2, 8–20.
        """
        template, interpolation_box = template.copy(), template.shape

        blob_window = window_blob(width=width)
        template = apply_window_filter(arr=template, filter_window=blob_window)

        if not np.all(template.shape == interpolation_box):
            template = self.interpolate_box(box=interpolation_box, arr=template)

        return template

    def hamming_filter(self, template: NDArray, width: int) -> NDArray:
        """
        Apply Hamming filter defined as:

        .. math::
            f_{hamming} = 0.54 - 0.46\\cos(\\frac{2\\pi n}{M-1})
            0 \\leq n \\leq M-1

        Parameters
        ----------
        template : NDArray
            The input atomic structure map.
        width : int
            Width of the filter window.

        Returns
        -------
        NDArray
            Simulated electron densities.
        """
        template, interpolation_box = template.copy(), template.shape

        hamming_window = np.hamming(int(width))
        hamming_window /= hamming_window.sum()

        template = apply_window_filter(arr=template, filter_window=hamming_window)

        if not np.all(template.shape == interpolation_box):
            template = self.interpolate_box(box=interpolation_box, arr=template)

        return template

    def rank_filter(self, template: NDArray, rank: int) -> NDArray:
        """
        Perform rank filtering.

        Parameters
        ----------
        template : NDArray
            The input atomic structure map.
        rank : int
            Footprint value. 0 -> minimum filter, -1 -> maximum filter.

        Returns
        -------
        NDArray
            Simulated electron densities.
        """
        if template is None:
            raise ValueError("Argument template missing")
        template = template.copy()
        interpolation_box = template.shape

        size = rank // 2
        if size <= 1:
            size = 3

        template = rank_filter(template, rank=rank, size=size)
        template = self.interpolate_box(box=interpolation_box, arr=template)

        return template

    def mipmap_filter(self, template: NDArray, level: int) -> NDArray:
        """
        Perform mip map antialiasing filtering.

        Parameters
        ----------
        template : NDArray
            The input atomic structure map.
        level : int
            Pyramid layer. Resolution decreases cubically with level.

        Returns
        -------
        NDArray
            Simulated electron densities.
        """
        array = template.copy()
        interpolation_box = array.shape

        print(array.shape)

        for k in range(template.ndim):
            array = decimate(array, q=level, axis=k)

        print(array.shape)
        template = zoom(array, np.divide(template.shape, array.shape))
        template = self.interpolate_box(box=interpolation_box, arr=template)

        return template

    def wavelet_filter(
        self,
        template: NDArray,
        level: int,
        wavelet: str = "bior2.2",
    ) -> NDArray:
        """
        Perform dyadic wavelet decomposition.

        Parameters
        ----------
        template : NDArray
            The input atomic structure map.
        level : int
            Scale of the wavelet transform.
        wavelet : str, optional
            Mother wavelet used for decomposition. Default is 'bior2.2'.

        Returns
        -------
        NDArray
            Simulated electron densities.
        """
        if wavelet not in wavelist(kind="discrete"):
            raise NotImplementedError(
                "Print argument wavelet has to be one of the following: %s",
                ", ".join(wavelist(kind="discrete")),
            )

        template, interpolation_box = template.copy(), template.shape
        decomp = wavedecn(template, level=level, wavelet=wavelet)

        for i in range(1, level + 1):
            decomp[i] = {k: np.zeros_like(v) for k, v in decomp[i].items()}

        template = waverecn(coeffs=decomp, wavelet=wavelet)
        template = self.interpolate_box(box=interpolation_box, arr=template)

        return template

    @staticmethod
    def molmap(
        coordinates: NDArray,
        weights: Tuple[float],
        resolution: float,
        sigma_factor: float = 1 / (np.pi * np.sqrt(2)),
        cutoff_value: float = 5.0,
        origin: Tuple[float] = None,
        shape: Tuple[int] = None,
        sampling_rate: float = None,
    ) -> NDArray:
        """
        Compute the electron densities analogous to Chimera's molmap function.

        Parameters
        ----------
        coordinates : NDArray
            A N x 3 array containing atomic coordinates in z, y, x format.
        weights : [float]
            The weights to use for the entries in coordinates.
        resolution : float
            The product of resolution and sigma_factor gives the sigma used to
            compute the discretized Gaussian.
        sigma_factor : float
            The factor used with resolution to compute sigma. Default is 1 / (π√2).
        cutoff_value : float
            The cutoff value for the Gaussian kernel. Default is 5.0.
        origin : (float,)
            The origin of the coordinate system used in coordinates. If not specified,
            the minimum coordinate along each axis is used.
        shape : (int,)
            The shape of the output array. If not specified, the function computes the
            smallest output array that contains all atoms.
        sampling_rate : float
            The Ångstrom per voxel of the output array. If not specified, the function
            sets this value to resolution/3.

        References
        ----------
        ..[1] https://www.cgl.ucsf.edu/chimera/docs/UsersGuide/midas/molmap.html

        Returns
        -------
        NDArray
            A numpy array containing the simulated electron densities.
        """
        if sampling_rate is None:
            sampling_rate = resolution * (1.0 / 3)

        coordinates = coordinates.copy()
        if origin is None:
            origin = coordinates.min(axis=0)
        if shape is None:
            positions = (coordinates - origin) / sampling_rate
            shape = positions.max(axis=0).astype(int)[::-1] + 2

        positions = (coordinates - origin) / sampling_rate
        positions = positions[:, ::-1]

        out = np.zeros(shape, dtype=np.float32)
        sigma = sigma_factor * resolution
        sigma_grid = sigma / sampling_rate
        sigma_grid2 = sigma_grid * sigma_grid
        for index, point in enumerate(np.rollaxis(positions, 0)):
            starts = np.maximum(np.ceil(point - cutoff_value * sigma_grid), 0).astype(
                int
            )
            stops = np.minimum(
                np.floor(point + cutoff_value * sigma_grid), shape
            ).astype(int)

            grid_index = np.meshgrid(
                *[range(start, stop) for start, stop in zip(starts, stops)]
            )
            distances = np.einsum(
                "aijk->ijk",
                np.array([(grid_index[i] - point[i]) ** 2 for i in range(len(point))]),
                dtype=np.float64,
            )
            np.add.at(
                out,
                tuple(grid_index),
                weights[index] * np.exp(-0.5 * distances / sigma_grid2),
            )

        out *= np.power(2 * np.pi, -1.5) * np.power(sigma, -3)
        return out

    def interpolate_box(
        self, arr: NDArray, box: Tuple[int], kind: str = "nearest"
    ) -> NDArray:
        """
        Resample ``arr`` within ``box`` using ``kind`` interpolation.

        Parameters
        ----------
        arr : NDArray
            The input numpy array.
        box : tuple of int
            Tuple of integers corresponding to the shape of the output array.
        kind : str, optional
            Interpolation method used (see scipy.interpolate.interp1d).
            Default is 'nearest'.

        Raises
        ------
        ValueError
            If the shape of box does not match arr.ndim

        Returns
        -------
        NDArray
            Interpolated numpy array.
        """
        if len(box) != arr.ndim:
            raise ValueError(f"Expected box of {arr.ndim}, got {len(box)}")

        for axis, size in enumerate(box):
            f = interp1d(
                np.linspace(0, 1, arr.shape[axis]),
                arr,
                kind=kind,
                axis=axis,
                fill_value="extrapolate",
            )
            arr = f(np.linspace(0, 1, size))

        return arr

    @staticmethod
    def fftfreqn(shape: NDArray, sampling_rate: NDArray) -> NDArray:
        """
        Calculate the N-dimensional equivalent to the inverse fftshifted
        absolute of numpy's fftfreq function, supporting anisotropic sampling.

        Parameters
        ----------
        shape : NDArray
            The shape of the N-dimensional array.
        sampling_rate : NDArray
            The sampling rate in the N-dimensional array.

        Returns
        -------
        NDArray
            A numpy array representing the norm of indices after normalization.

        Examples
        --------
        >>> import numpy as np
        >>> from dge import Preprocessor
        >>> freq = Preprocessor().fftfreqn((10,), 1)
        >>> freq_numpy = np.fft.fftfreq(10, 1)
        >>> np.allclose(freq, np.abs(np.fft.ifftshift(freq_numpy)))
        """
        indices = np.indices(shape).T
        norm = np.multiply(shape, sampling_rate)
        indices -= np.divide(shape, 2).astype(int)
        indices = np.divide(indices, norm)
        return np.linalg.norm(indices, axis=-1).T

    def _approximate_butterworth(
        self,
        radial_frequencies: NDArray,
        lowcut: float,
        highcut: float,
        gaussian_sigma: float,
    ) -> NDArray:
        """
        Approximate a Butterworth band-pass filter for given radial frequencies.
        The DC component of the filter is at the origin.

        Parameters
        ----------
        radial_frequencies : NDArray
            The radial frequencies for which the Butterworth band-pass
            filter is to be calculated.
        lowcut : float
            The lower cutoff frequency for the band-pass filter.
        highcut : float
            The upper cutoff frequency for the band-pass filter.
        gaussian_sigma : float
            The sigma value for the Gaussian smoothing applied to the filter.

        Returns
        -------
        NDArray
            A numpy array representing the approximate Butterworth
            band-pass filter applied to the radial frequencies.
        """
        bpf = ((radial_frequencies <= highcut) & (radial_frequencies >= lowcut)) * 1.0
        bpf = self.gaussian_filter(template=bpf, sigma=gaussian_sigma, fourier=False)
        bpf[bpf < np.exp(-2)] = 0
        bpf = np.fft.ifftshift(bpf)

        return bpf

    def bandpass_filter(
        self,
        template: NDArray,
        minimum_frequency: float,
        maximum_frequency: float,
        sampling_rate: NDArray = None,
        gaussian_sigma: float = 0.0,
    ) -> NDArray:
        """
        Apply a band-pass filter on the provided template, using a
        Butterworth approximation.

        Parameters
        ----------
        template : NDArray
            The input numpy array on which the band-pass filter should be applied.
        minimum_frequency : float
            The lower boundary of the frequency range to be preserved. Lower values will
            retain broader, more global features.
        maximum_frequency : float
            The upper boundary of the frequency range to be preserved.  Higher values
            will emphasize finer details and potentially noise.

        sampling_rate : NDarray, optional
            The sampling rate along each dimension.
        gaussian_sigma : float, optional
            Sigma value for the gaussian smoothing to be applied to the filter.

        Returns
        -------
        NDArray
            Bandpass filtered numpy array.
        """
        bpf = self.bandpass_mask(
            shape=template.shape,
            minimum_frequency=minimum_frequency,
            maximum_frequency=maximum_frequency,
            sampling_rate=sampling_rate,
            gaussian_sigma=gaussian_sigma,
            omit_negative_frequencies=False,
        )

        fft_data = np.fft.fftn(template)
        np.multiply(fft_data, bpf, out=fft_data)
        ret = np.real(np.fft.ifftn(fft_data))
        return ret

    def bandpass_mask(
        self,
        shape: Tuple[int],
        minimum_frequency: float,
        maximum_frequency: float,
        sampling_rate: NDArray = None,
        gaussian_sigma: float = 0.0,
        omit_negative_frequencies: bool = True,
    ) -> NDArray:
        """
        Compute an approximate Butterworth bundpass filter. The returned filter
        has it's DC component at the origin.

        Parameters
        ----------
        shape : tuple of ints
            Shape of the returned bandpass filter.
        minimum_frequency : float
            The lower boundary of the frequency range to be preserved. Lower values will
            retain broader, more global features.
        maximum_frequency : float
            The upper boundary of the frequency range to be preserved.  Higher values
            will emphasize finer details and potentially noise.
        sampling_rate : NDarray, optional
            The sampling rate along each dimension.
        gaussian_sigma : float, optional
            Sigma value for the gaussian smoothing to be applied to the filter.
        omit_negative_frequencies : bool, optional
            Whether the wedge mask should omit negative frequencies, i.e. be
            applicable to non hermitian-symmetric fourier transforms.

        Returns
        -------
        NDArray
            Bandpass filtered.
        """
        if sampling_rate is None:
            sampling_rate = np.ones(len(shape))
        sampling_rate = np.asarray(sampling_rate, dtype=np.float32)
        sampling_rate /= sampling_rate.max()

        if minimum_frequency > maximum_frequency:
            minimum_frequency, maximum_frequency = maximum_frequency, minimum_frequency

        radial_freq = self.fftfreqn(shape, sampling_rate)
        bpf = self._approximate_butterworth(
            radial_frequencies=radial_freq,
            lowcut=minimum_frequency,
            highcut=maximum_frequency,
            gaussian_sigma=gaussian_sigma,
        )

        if omit_negative_frequencies:
            stop = 1 + (shape[-1] // 2)
            bpf = bpf[..., :stop]

        return bpf

    def wedge_mask(
        self,
        shape: Tuple[int],
        tilt_angles: NDArray,
        sigma: float = 0,
        omit_negative_frequencies: bool = True,
    ) -> NDArray:
        """
        Create a wedge mask with the same shape as template by rotating a
        plane according to tilt angles. The DC component of the filter is at the origin.

        Parameters
        ----------
        shape : Tuple of ints
            Shape of the output wedge array.
        tilt_angles : NDArray
            Tilt angles in format d dimensions N tilts [d x N].
        sigma : float, optional
            Standard deviation for Gaussian kernel used for smoothing the wedge.
        omit_negative_frequencies : bool, optional
            Whether the wedge mask should omit negative frequencies, i.e. be
            applicable to non hermitian-symmetric fourier transforms.

        Returns
        -------
        NDArray
            A numpy array containing the wedge mask.

        Notes
        -----
        The axis perpendicular to the tilts is the leftmost closest axis
        with minimal tilt.

        Examples
        --------
        >>> import numpy as np
        >>> from tme import Preprocessor
        >>> angles = np.zeros((3, 10))
        >>> angles[2, :] = np.linspace(-50, 55, 10)
        >>> wedge = Preprocessor().wedge_mask(
        >>>    shape = (50,50,50),
        >>>    tilt_angles = angles,
        >>>    omit_negative_frequencies = True
        >>>    )
        >>> wedge = np.fft.fftshift(wedge)

        This will create a wedge that is open along axis 1, tilted
        around axis 2 and propagated along axis 0. The code above would
        be equivalent to the following

        >>> wedge = Preprocessor().continuous_wedge_mask(
        >>>    shape = (50,50,50),
        >>>    start_tilt = 50,
        >>>    stop_tilt=55,
        >>>    tilt_axis=1,
        >>>    omit_negative_frequencies=False,
        >>>    infinite_plane=False
        >>>    )
        >>> wedge = np.fft.fftshift(wedge)

        with the difference being that :py:meth:`Preprocessor.continuous_wedge_mask`
        does not consider individual plane tilts.

        See Also
        --------
        :py:meth:`Preprocessor.continuous_wedge_mask`
        """
        plane = np.zeros(shape, dtype=np.float32)
        opening_axis = np.argmax(np.abs(tilt_angles), axis=0)
        slices = tuple(slice(a, a + 1) for a in np.divide(shape, 2).astype(int))
        plane_rotated = np.zeros_like(plane)
        wedge_volume = np.zeros_like(plane)
        for index in range(tilt_angles.shape[1]):
            potential_axes, *_ = np.where(
                np.abs(tilt_angles[:, index]) == np.abs(tilt_angles[:, index]).min()
            )
            largest_tilt = np.argmax(np.abs(tilt_angles[:, index]))
            opening_axis_index = np.argmin(np.abs(potential_axes - largest_tilt))
            opening_axis = potential_axes[opening_axis_index]
            rotation_matrix = euler_to_rotationmatrix(tilt_angles[:, index])
            plane_rotated.fill(0)
            plane.fill(0)
            subset = tuple(
                slice(None) if i != opening_axis else slices[opening_axis]
                for i in range(plane.ndim)
            )
            plane[subset] = 1
            Density.rotate_array(
                arr=plane,
                rotation_matrix=rotation_matrix,
                out=plane_rotated,
                use_geometric_center=True,
                order=1,
            )
            wedge_volume += plane_rotated

        wedge_volume = self.gaussian_filter(
            template=wedge_volume, sigma=sigma, fourier=False
        )
        wedge_volume = np.where(wedge_volume > np.exp(-2), 1, 0)
        wedge_volume = np.fft.ifftshift(wedge_volume)

        if omit_negative_frequencies:
            stop = 1 + (wedge_volume.shape[-1] // 2)
            wedge_volume = wedge_volume[..., :stop]

        return wedge_volume

    def continuous_wedge_mask(
        self,
        start_tilt: float,
        stop_tilt: float,
        shape: Tuple[int],
        tilt_axis: int = 1,
        sigma: float = 0,
        extrude_plane: bool = True,
        infinite_plane: bool = True,
        omit_negative_frequencies: bool = True,
    ) -> NDArray:
        """
        Generate a wedge in a given shape based on specified tilt angles and axis.
        The DC component of the filter is at the origin.

        Parameters
        ----------
        start_tilt : float
            Starting tilt angle in degrees, e.g. a stage tilt of 70 degrees
            would yield a start_tilt value of 70.
        stop_tilt : float
            Ending tilt angle in degrees, , e.g. a stage tilt of -70 degrees
            would yield a stop_tilt value of 70.
        tilt_axis : int
            Axis that runs through the empty part of the wedge.
            - 0 for X-axis
            - 1 for Y-axis
            - 2 for Z-axis
        shape : Tuple of ints
            Shape of the output wedge array.
        sigma : float, optional
            Standard deviation for Gaussian kernel used for smoothing the wedge.
        extrude_plane : bool, optional
            Whether the tilted plane is extruded to 3D. By default, this represents
            the effect of rotating a plane in 3D yielding a cylinder with wedge
            insertion. If set to False, the returned mask has spherical shape,
            analogous to rotating a line in 3D.
        omit_negative_frequencies : bool, optional
            Whether the wedge mask should omit negative frequencies, i.e. be
            applicable to non hermitian-symmetric fourier transforms.
        infinite_plane : bool, optional
            Whether the plane should be considered to be larger than the shape. In this
            case the output wedge mask fill have no spheric component.

        Returns
        -------
        NDArray
            Array of the specified shape with the wedge created based on
            the tilt angles.

        Examples
        --------
        >>> wedge = create_wedge(30, 60, 1, (64, 64, 64))

        Notes
        -----
        The rotation plane is spanned by the tilt axis and the leftmost dimension
        that is not the tilt axis.

        See Also
        --------
        :py:meth:`Preprocessor.wedge_mask`
        """
        shape_center = np.divide(shape, 2).astype(int)

        opening_axis = tilt_axis
        base_axis = (tilt_axis + 1) % len(shape)

        grid = (np.indices(shape).T - shape_center).T

        start_radians = np.tan(np.radians(90 - start_tilt))
        stop_radians = np.tan(np.radians(-1 * (90 - stop_tilt)))
        max_tan_value = np.tan(np.radians(90)) + 1

        with np.errstate(divide="ignore", invalid="ignore"):
            ratios = np.where(
                grid[opening_axis] == 0,
                max_tan_value,
                grid[base_axis] / grid[opening_axis],
            )

        wedge = np.logical_or(start_radians <= ratios, stop_radians >= ratios).astype(
            np.float32
        )

        if extrude_plane:
            distances = np.sqrt(grid[base_axis] ** 2 + grid[opening_axis] ** 2)
        else:
            distances = np.linalg.norm(grid, axis=0)

        if not infinite_plane:
            np.multiply(wedge, distances <= shape[opening_axis] // 2, out=wedge)

        wedge = self.gaussian_filter(template=wedge, sigma=sigma, fourier=False)
        wedge = np.fft.ifftshift(wedge > np.exp(-2))

        if omit_negative_frequencies:
            stop = 1 + (wedge.shape[-1] // 2)
            wedge = wedge[..., :stop]

        return wedge

    @staticmethod
    def _fourier_crop_mask(old_shape: NDArray, new_shape: NDArray) -> NDArray:
        """
        Generate a mask for Fourier cropping.

        Parameters
        ----------
        old_shape : NDArray
            The original shape of the array before cropping.
        new_shape : NDArray
            The new desired shape for the array after cropping.

        Returns
        -------
        NDArray
            The mask array for Fourier cropping.
        """
        mask = np.zeros(old_shape, dtype=bool)
        mask[tuple(np.indices(new_shape))] = 1
        box_shift = np.floor(np.divide(new_shape, 2)).astype(int)
        mask = np.roll(mask, shift=-box_shift, axis=range(len(old_shape)))
        return mask

    def fourier_crop(
        self,
        template: NDArray,
        reciprocal_template_filter: NDArray,
        crop_factor: float = 3 / 2,
    ) -> NDArray:
        """
        Perform Fourier uncropping on a given template.

        Parameters
        ----------
        template : NDArray
            The original template to be uncropped.
        reciprocal_template_filter : NDArray
            The filter to be applied in the Fourier space.
        crop_factor : float
            Cropping factor over reeciprocal_template_filter boundary.

        Returns
        -------
        NDArray
            The uncropped template.
        """
        new_boxsize = np.zeros(template.ndim, dtype=int)
        for i in range(template.ndim):
            slices = tuple(
                slice(0, 1) if j != i else slice(template.shape[i] // 2)
                for j in range(template.ndim)
            )
            filt = np.squeeze(reciprocal_template_filter[slices])
            new_boxsize[i] = np.ceil((np.max(np.where(filt > 0)) + 1) * crop_factor) * 2

        if np.any(np.greater(new_boxsize, template.shape)):
            new_boxsize = np.array(template.shape).copy()

        mask = self._fourier_crop_mask(old_shape=template.shape, new_shape=new_boxsize)
        arr_ft = np.fft.fftn(template)
        arr_ft *= np.prod(new_boxsize) / np.prod(template.shape)
        arr_ft = np.reshape(arr_ft[mask], new_boxsize)
        arr_cropped = np.real(np.fft.ifftn(arr_ft))
        return arr_cropped

    def fourier_uncrop(
        self, template: NDArray, reciprocal_template_filter: NDArray
    ) -> NDArray:
        """
        Perform an uncrop operation in the Fourier space.

        Parameters
        ----------
        template : NDArray
            The input array.
        reciprocal_template_filter : NDArray
            The filter to be applied in the Fourier space.

        Returns
        -------
        NDArray
            Uncropped template with shape reciprocal_template_filter.
        """
        mask = self._fourier_crop_mask(
            old_shape=reciprocal_template_filter.shape, new_shape=template.shape
        )
        ft_vol = np.zeros_like(mask)
        ft_vol[mask] = np.fft.fftn(template).ravel()
        ft_vol *= np.divide(np.prod(mask.shape), np.prod(template.shape)).astype(
            ft_vol.dtype
        )
        reciprocal_template_filter = reciprocal_template_filter.astype(ft_vol.dtype)
        np.multiply(ft_vol, reciprocal_template_filter, out=ft_vol)
        ret = np.real(np.fft.ifftn(ft_vol))
        return ret
